#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Prepared functions for scheduling dispatcher.

Dispatcher strategy configuration functions. All of them are receiving a dict
as a first argument containing:

  * previous_retry - value of previous retry, None if scheduling for the first time
  * active_nodes - active nodes within the system
  * failed_nodes - failed nodes within the system
  * new_started_nodes - newly started nodes in the system, these nodes are already in active_nodes
  * new_fallback_nodes - newly started fallback nodes, these nodes are already in active_nodes
  * finished nodes - already finished nodes

Other parameters are specific for strategy function and are configurable from YAML configuration file.
"""

from random import randint as gen_random

# There are done checks on user-defined strategies, so keep args
# pylint: disable=unused-argument


def linear_increase(status, start_retry, max_retry, step):
    """Increase linearly if no node started, decrease drastically to start_retry if no node scheduled.

    :param status: flow status dict
    :param start_retry: starting retry to use
    :param max_retry: upper limit of scheduling
    :param step: step to use in linear increase
    """
    if not status['active_nodes']:
        return None

    if status['previous_retry'] is None:
        return start_retry

    if status['new_started_nodes'] or status['new_fallback_nodes']:
        retry = status['previous_retry'] + step
        return retry if retry < max_retry else max_retry

    return start_retry


def linear_adapt(status, start_retry, max_retry, step):
    """Increase linearly if no node started, decrease linearly if a node scheduled.

    :param status: flow status dict
    :param start_retry: starting retry to use
    :param max_retry: upper limit of scheduling
    :param step: step to use in linear increase
    """
    if not status['active_nodes']:
        return None

    if status['previous_retry'] is None:
        return start_retry

    if status['new_started_nodes'] or status['new_fallback_nodes']:
        retry = status['previous_retry'] + step
        return retry if retry < max_retry else max_retry

    retry = status['previous_retry'] - step
    return retry if retry > start_retry else start_retry


def biexponential_increase(status, start_retry, max_retry):
    """Increase exponentially if no node started, decrease drastically to start_retry if a node scheduled.

    :param status: flow status dict
    :param start_retry: starting retry to use
    :param max_retry: upper limit of scheduling
    """
    if not status['active_nodes']:
        return None

    if status['previous_retry'] is None:
        return start_retry

    if status['new_started_nodes'] or status['new_fallback_nodes']:
        retry = status['previous_retry'] * 2
        return retry if retry < max_retry else max_retry

    return start_retry


def biexponential_decrease(status, start_retry, stop_retry):
    """Decrease by div 2 each time if no node started, decrease to stop_retry if a node scheduled.

    :param status: flow status dict
    :param start_retry: starting retry to use
    :param stop_retry: upper limit of scheduling
    """
    if not status['active_nodes']:
        return None

    if status['previous_retry'] is None:
        return start_retry

    retry = status['previous_retry'] / 2
    return retry if retry > stop_retry else stop_retry


def biexponential_adapt(status, start_retry, max_retry):
    """Increase exponentially if no node started, decrease exponentially if a node scheduled.

    :param status: flow status dict
    :param start_retry: starting retry to use
    :param max_retry: upper limit of scheduling
    """
    if status['previous_retry'] is None:
        return start_retry

    if not status['active_nodes']:
        return None

    if status['new_started_nodes'] or status['new_fallback_nodes']:
        retry = status['previous_retry'] * 2
        return retry if retry < max_retry else max_retry

    retry = status['previous_retry'] / 2
    return retry if retry > start_retry else start_retry


def random(status, start_retry, max_retry):
    """Schedule randomly.

    :param status: flow status dict
    :param start_retry: lower limit of scheduling
    :param max_retry: upper limit of scheduling
    """
    if not status['active_nodes']:
        return None

    return gen_random(start_retry, max_retry)  # nosec


def constant(status, retry):
    """Schedule constantly.

    :param status: flow status dict
    :param retry: constant retry timeout
    """
    if not status['active_nodes']:
        return None

    return retry



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""User configuration."""

import yaml

from selinon.errors import RequestError


class _ConfigSingleton(type):
    """Config singleton metaclass."""

    _instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(_ConfigSingleton, cls).__call__(cls._config, *args, **kwargs)
        return cls._instance

    @classmethod
    def set_config(mcs, config_path):
        """Set config which should be used in within Config singleton.

        :param config_path: a path to configuration that should be used
        :type config_path: str
        """
        # set _config before the singleton is instantiated
        assert mcs._instance is None  # nosec
        mcs._config = config_path


def style_configuration(style_name):
    """Adjust style based on default style and provided style configuration (if any)."""
    def decorator(func):
        """Decorate a function call."""
        def wrapper(config_instance):
            """Wrap style configuration that makes sure style is adjusted based on user configuration."""
            property_name = '_' + func.__name__
            style_value = getattr(config_instance, property_name)

            if not style_value:
                style_value = {}
                # Apply the default style first
                default_style_value = getattr(config_instance, 'DEFAULT_' + func.__name__.upper())
                style_value.update(default_style_value)
                # Overwrite styles based on user's configuration
                # Style name could be directly determined from function name, but leave it this way for the
                # sake of readability
                config_style_value = (config_instance.raw_config.get('style') or {}).get(style_name, {})
                style_value.update(config_style_value)
                setattr(config_instance, property_name, style_value)

            return func(config_instance)

        return wrapper

    return decorator


class UserConfig(metaclass=_ConfigSingleton):
    # pylint: disable=too-many-instance-attributes
    """Configuration supplied by user."""

    DEFAULT_STYLE_TASK = {
        'style': 'filled',
        'color': 'black',
        'fillcolor': '#66cfa7',
        'shape': 'ellipse'
    }
    DEFAULT_STYLE_FLOW = {
        'style': 'filled',
        'color': 'black',
        'fillcolor': '#197a9f',
        'shape': 'box3d'
    }
    DEFAULT_STYLE_CONDITION = {
        'style': 'filled',
        'color': 'gray',
        'fillcolor': '#e8e3c8',
        'shape': 'octagon'
    }
    DEFAULT_STYLE_CONDITION_FOREACH = {
        'style': 'filled',
        'color': 'gray',
        'fillcolor': '#e8e3c8',
        'shape': 'doubleoctagon'
    }
    DEFAULT_STYLE_STORAGE = {
        'style': 'filled',
        'color': 'black',
        'fillcolor': '#894830',
        'shape': 'cylinder'
    }
    DEFAULT_STYLE_EDGE = {
        'arrowType': 'open',
        'color': 'black'
    }
    DEFAULT_STYLE_STORE_EDGE = {
        'arrowType': 'open',
        'color': '#894830',
        'style': 'dashed'
    }
    DEFAULT_STYLE_GRAPH = {
    }
    DEFAULT_STYLE_FALLBACK_EDGE = {
        'arrowType': 'open',
        'color': '#cc1010'
    }
    DEFAULT_STYLE_FALLBACK_TRUE = {
        'style': 'filled',
        'color': 'black',
        'fillcolor': '#5af47b',
        'shape': 'plain'
    }

    def __init__(self, config=None):
        """Instantiate configuration."""
        self.raw_config = {}
        self.config_path = config

        # These get assigned in style_configuration decorator.
        self._style_task = None
        self._style_flow = None
        self._style_condition = None
        self._style_condition_foreach = None
        self._style_storage = None
        self._style_edge = None
        self._style_store_edge = None
        self._style_graph = None
        self._style_fallback_edge = None
        self._style_fallback_node = None
        self._style_fallback_true = None

        if self.config_path is not None:
            try:
                with open(self.config_path) as input_file:
                    self.raw_config = yaml.load(input_file, Loader=yaml.SafeLoader)
            except Exception as exc:
                raise RequestError("Unable to load or parse style configuration file %r: %s"
                                   % (self.config_path, str(exc))) from exc

    @style_configuration('task')
    def style_task(self):
        """Return style for tasks in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        return self._style_task

    @style_configuration('flow')
    def style_flow(self):
        """Return style for a flow node in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        return self._style_flow

    @style_configuration('condition')
    def style_condition(self):
        """Return style for conditions in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        return self._style_condition

    @style_configuration('condition_foreach')
    def style_condition_foreach(self):
        """Return style for foreach edges in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        return self._style_condition_foreach

    @style_configuration('storage')
    def style_storage(self):
        """Return style for storage in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        return self._style_storage

    @style_configuration('edge')
    def style_edge(self):
        """Return style for edges in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        return self._style_edge

    @style_configuration('store_edge')
    def style_store_edge(self):
        """Return style for edges that lead to a storage in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        return self._style_store_edge

    @style_configuration('graph')
    def style_graph(self):
        """Return style for the whole graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        return self._style_graph

    @style_configuration('fallback_edge')
    def style_fallback_edge(self):
        """Return style for fallback edges.

        :return: style definition
        :rtype: dict
        """
        return self._style_fallback_edge

    @style_configuration('fallback_true')
    def style_fallback_true(self):
        """Return style for fallback true node.

        :return: style definition
        :rtype: dict
        """
        return self._style_fallback_true



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Selinon SQL Database adapter - PostgreSQL."""

import os
from selinon.data_storage import SelinonMissingDataException
from selinon import DataStorage

try:
    from sqlalchemy import create_engine
    from sqlalchemy.orm import sessionmaker
    from sqlalchemy_utils import create_database
    from sqlalchemy_utils import database_exists
except ImportError as exc:
    raise ImportError("Please install dependencies using `pip3 install selinon[postgresql]`") from exc

from .models import Result


class PostgreSQL(DataStorage):
    """Selinon SQL Database adapter - PostgreSQL."""

    def __init__(self, connection_string, encoding='utf-8', echo=False):
        """Initialize PostgreSQL adapter from YAML configuration file.

        :param connection_string: connection string to be used to connect to PostgreSQL to
        :param encoding: encoding to be used
        :param echo: perform echo on queries
        """
        super().__init__()

        self.engine = create_engine(connection_string.format(**os.environ), encoding=encoding, echo=echo)
        self.session = None

    def is_connected(self):  # noqa
        return self.session is not None

    def connect(self):  # noqa
        if not database_exists(self.engine.url):
            create_database(self.engine.url)

        self.session = sessionmaker(bind=self.engine)()
        Result.metadata.create_all(self.engine)

    def disconnect(self):  # noqa
        if self.is_connected():
            self.session.close()
            self.session = None

    def retrieve(self, flow_name, task_name, task_id):  # noqa
        assert self.is_connected()  # nosec

        record = self.session.query(Result).filter_by(task_id=task_id).one()

        assert record.task_name == task_name  # nosec
        return record.result

    def store(self, node_args, flow_name, task_name, task_id, result):  # noqa
        assert self.is_connected()  # nosec

        record = Result(node_args, flow_name, task_name, task_id, result)
        try:
            self.session.add(record)
            self.session.commit()
        except Exception:
            self.session.rollback()
            raise

        return record.id

    def store_error(self, node_args, flow_name, task_name, task_id, exc_info):  # noqa
        # just to make pylint happy
        raise NotImplementedError()

    def delete(self, flow_name, task_name, task_id):  # noqa
        assert self.is_connected()  # nosec

        response = self.session.query(Result).filter_by(task_id=task_id).delete()

        if response == 0:
            raise SelinonMissingDataException("Record not found")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import time
import datetime
import flexmock
from selinon_test_case import SelinonTestCase
from selinon import SystemState


class TestFlow(SelinonTestCase):
    def test_storage_issues(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import pytest
from flexmock import flexmock
from selinon import SelinonTask
from selinon import StoragePool
from selinon_test_case import SelinonTestCase
from selinon import NoParentNodeError
from selinon import Retry


class _MyTask(SelinonTask):
    def run(self, node_args):
        super().run(node_args)


_MYTASK_PARAMS = {
    'task_name': 'task1',
    'flow_name': 'flow1',
    'parent': {
        'task2': '<task2-id>',
        'flow2': {
            'task3': ['<flow2-task2-id>']
        }
    },
    'task_id': '<task-id>',
    'dispatcher_id': '<dispatcher-id>'
}


task_test = [
    (_MyTask(**_MYTASK_PARAMS), _MYTASK_PARAMS)
]


@pytest.mark.parametrize("task, params", task_test)
class TestSelinonTask(SelinonTestCase):
    def test_init(self, task, params):
        assert task.task_name == params['task_name']
        assert task.flow_name == params['flow_name']
        assert task.parent == params['parent']
        assert task.dispatcher_id == params['dispatcher_id']

    def test_parent_task_result(self, task, params):
        parent_task_name = 'task2'
        parent_task_id = params['parent'][parent_task_name]
        result = 'foo'

        flexmock(StoragePool)\
            .should_receive('retrieve')\
            .with_args(params['flow_name'], parent_task_name, parent_task_id)\
            .and_return(result)

        assert task.parent_task_result(parent_task_name) == result

    def test_parent_task_result_error(self, task, params):
        with pytest.raises(NoParentNodeError):
            task.parent_task_result('some-not-existing-task')

    def test_parent_flow_result(self, task, params):
        parent_flow_name = 'flow2'
        parent_task_name = 'task3'
        index = 0
        parent_task_id = params['parent'][parent_flow_name][parent_task_name][index]
        result = 'foo'

        flexmock(StoragePool)\
            .should_receive('retrieve')\
            .with_args(parent_flow_name, parent_task_name, parent_task_id)\
            .and_return(result)

        assert task.parent_flow_result(parent_flow_name, parent_task_name, index=0) == result

    def test_parent_flow_result_error_no_flow(self, task, params):
        with pytest.raises(NoParentNodeError):
            task.parent_flow_result('some-not-existing-flow', 'task2', index=0)

    def test_parent_flow_result_error_no_task(self, task, params):
        with pytest.raises(NoParentNodeError):
            task.parent_flow_result('flow2', 'some-not-existing-task', index=0)

    def test_parent_flow_result_error_index(self, task, params):
        with pytest.raises(NoParentNodeError):
            task.parent_flow_result('flow2', 'task3', index=42)

    def test_storage(self, task, params):
        storage_instance = 'some-storage'

        flexmock(StoragePool)\
            .should_receive('get_storage_by_task_name')\
            .with_args(params['task_name'])\
            .and_return(storage_instance)

        assert task.storage == storage_instance

    def test_retry(self, task, params):
        with pytest.raises(Retry):
            task.retry()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import os
from get_task_instance import GetTaskInstance
from queue_mock import QueueMock
from strategy_mock import strategy_function
from storage_task_name_mock import StorageTaskNameMock

from celery.result import AsyncResult
from selinon.caches import LRU
from selinon.config import Config
from selinon.system_state import SystemState
from selinon.trace import Trace


class _ThrottleTasks:
    def __init__(self, is_flow, throttle_conf=None):
        self._is_flow = is_flow
        self._throttle_conf = throttle_conf or {}

    def __getitem__(self, item):
        if self._is_flow(item):
            raise KeyError("Unable to find throttle flow configuration in task throttle configuration")
        return self._throttle_conf.get(item)


class _AsyncResultCacheMock:
    def __init__(self, is_flow):
        self.is_flow = is_flow
        self.cache = LRU(max_cache_size=0)

    def __getitem__(self, item):
        if not self.is_flow(item):
            raise KeyError("Unable to subscribe for AsyncResult cache since '%s' is not a flow", item)
        return self.cache


class _TaskResultCacheMock:
    def __init__(self):
        self.cache = LRU(max_cache_size=0)

    def __getitem__(self, item):
        return self.cache


class _SelectiveRunFunctionMock:
    def __getitem__(self, item):
        def func(flow_name, node_name, node_args, task_names, storage_pool):
            # Always return None so we always execute task before the desired one
            return None
        return func


class SelinonTestCase:
    """
    Main class for Selinon testing
    """
    # dir for data files for tests
    DATA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')

    def setup_method(self, method):
        """Clean up all class attributes from previous runs."""
        AsyncResult.clear()
        GetTaskInstance.clear()
        SystemState._throttled_tasks = {}
        SystemState._throttled_flows = {}
        # Make sure we restore tracing function in tests
        Trace._trace_functions = []

    def teardown_method(self, method):
        """Clean up resources and configuration after a test."""
        Config.initialized = False
        Config.migration_dir = None

    def init(self, edge_table, **kwargs):
        """
        :param edge_table: table dict as defined in YAML file containing edges
        :param kwargs: additional parameters for test configuration
        """
        Config.edge_table = edge_table
        flows = list(edge_table.keys())

        Config.flows = kwargs.pop('flows', flows)
        Config.nowait_nodes = kwargs.pop('nowait_nodes', dict.fromkeys(flows, []))
        Config.eager_failures = kwargs.pop('eager_failures', dict.fromkeys(flows, []))
        Config.get_task_instance = kwargs.pop('get_task_instance', GetTaskInstance())
        Config.failures = kwargs.pop('failures', {})
        Config.propagate_node_args = kwargs.pop('propagate_node_args', dict.fromkeys(flows, False))
        Config.propagate_parent = kwargs.pop('propagate_parent', dict.fromkeys(flows, False))
        Config.propagate_finished = kwargs.pop('propagate_finished', dict.fromkeys(flows, False))
        Config.propagate_compound_finished = kwargs.pop('propagate_compound_finished', dict.fromkeys(flows, False))
        Config.max_retry = kwargs.pop('max_retry', {})
        Config.retry_countdown = kwargs.pop('retry_countdown', {})
        Config.task_queues = kwargs.pop('task_queues', QueueMock())
        Config.dispatcher_queues = kwargs.pop('dispatcher_queues', QueueMock())
        Config.strategies = kwargs.pop('strategies', dict.fromkeys(flows, strategy_function))
        # TODO: this is currently unused as we do not have tests for store()
        Config.storage_readonly = kwargs.pop('storage_readonly', {})
        Config.storage_task_name = kwargs.pop('storage_task_name', StorageTaskNameMock())
        Config.task2storage_mapping = kwargs.pop('task2storage_mapping', {})
        Config.storage2storage_cache = kwargs.pop('storage2storage_cache', _TaskResultCacheMock())
        Config.node_args_from_first = kwargs.pop('node_args_from_first', dict.fromkeys(flows, False))
        Config.throttle_flows = kwargs.pop('throttle_flows', dict.fromkeys(flows, None))
        Config.throttle_tasks = kwargs.pop('throttle_tasks', _ThrottleTasks(Config.is_flow,
                                                                            kwargs.get('throttle_tasks_conf')))
        Config.storage_mapping = kwargs.pop('storage_mapping', {})
        Config.output_schemas = kwargs.pop('output_schemas', {})
        Config.async_result_cache = kwargs.pop('async_result_cache', _AsyncResultCacheMock(Config.is_flow))
        Config.selective_run_task = kwargs.pop('selective_run_task', _SelectiveRunFunctionMock())
        Config.initialized = True

        if kwargs:
            raise ValueError("Unknown config options provided: %s" % set(kwargs.keys()))

        self._update_edge_table()

    @staticmethod
    def _update_edge_table():
        """
        As edge table stores some metadata for debugging, let's fake them during testing
        """
        for key in Config.edge_table.keys():
            for entry in Config.edge_table[key]:
                entry['foreach_str'] = 'foreach_str'
                entry['condition_str'] = 'condition_str'

    @staticmethod
    def cond_true(db, node_args):
        """
        Condition for edge transition

        :return: always true
        """
        return True

    @staticmethod
    def cond_false(db, node_args):
        """
        Condition for edge transition

        :return: always false
        """
        return False

    @property
    def instantiated_tasks(self):
        """
        :return: list of all instantiated tasks
        """
        return Config.get_task_instance.tasks

    @property
    def instantiated_flows(self):
        """
        :return: list of all instantiated flows
        """
        return Config.get_task_instance.flows

    @staticmethod
    def set_finished(node, result=None):
        """
        Mark node (task or flow) as successfully finished

        :param node: node to mark as finished
        :param result: result of node, if None, result is not set
        """
        AsyncResult.set_finished(node.task_id)
        if result is not None:
            AsyncResult.set_result(node.task_id, result)

    @staticmethod
    def set_failed(node, exc=None):
        """
        Mark node (task or flow) as failed

        :param node: node to mark as failed
        :param result: exception (result for Celery) that was raised
        """
        AsyncResult.set_failed(node.task_id)
        AsyncResult.set_result(node.task_id, exc or ValueError("Some unexpected exception in node"))

    @staticmethod
    def get_task(task_name, idx=None):
        """
        Get task by its name

        :param task_name: name of task
        :param idx: index of task in instantiated tasks, if there are more tasks of type, the last is taken
        :return: task
        """
        tasks = Config.get_task_instance.task_by_name(task_name)
        return tasks[idx if idx is not None else -1]

    @staticmethod
    def get_all_tasks(task_name):
        """
        Get all tasks by task name

        :param task_name: name of task
        :return: tasks
        """
        return Config.get_task_instance.task_by_name(task_name)

    @staticmethod
    def get_flow(flow_name, idx=None):
        """
        Get flow by its name

        :param flow_name: name of flow
        :param idx: index of flow in instantiated flows, if there are more tasks of type, the last is taken
        :return: flow
        """
        tasks = Config.get_task_instance.flow_by_name(flow_name)
        return tasks[idx or -1]

    @staticmethod
    def get_all_flows(flow_name):
        """
        Get all flows by its name

        :param flow_name: name of flow
        :return: flows
        """
        return Config.get_task_instance.flow_by_name(flow_name)

    @property
    def get_task_instance(self):
        """
        :return: GetTaskInstance in the current context
        """
        return Config.get_task_instance

    @staticmethod
    def remove_all_flows_by_name(flow_name):
        """ Remove all flows with the given name """
        Config.get_task_instance.remove_all_flows_by_name(flow_name)

    @staticmethod
    def remove_all_tasks_by_name(task_name):
        """ Remove all flows with the given name """
        Config.get_task_instance.remove_all_tasks_by_name(task_name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################


class Task:
    def __init__(self):
        # In case of instantiating SelinonTaskEnvelope or Dispatcher, we are not passing any arguments
        self.task_name = None
        self.flow_name = None
        self.parent = None
        self.node_args = None
        self.retried_count = None
        self.queue = None
        self.countdown = None
        self.dispatcher_id = None
        self.selective = None

    @property
    def task_id(self):
        return id(self)

    def apply_async(self, kwargs, queue, countdown=None):
        from selinon.config import Config

        # Ensure that SelinonTaskEnvelope kept parameters consistent
        self.flow_name = kwargs['flow_name']
        self.node_args = kwargs.get('node_args')
        self.parent = kwargs.get('parent')
        self.dispatcher_id = kwargs.get('dispatcher_id')

        # None if we have flow
        self.task_name = kwargs.get('task_name')
        self.retried_count = kwargs.get('retried_count')
        self.countdown = countdown
        self.selective = kwargs.get('selective')

        self.queue = queue
        Config.get_task_instance.register_node(self)

        return self

    @staticmethod
    def retry(exc, max_retry):
        # ensure that we are raising with max_retry equal to 0 so we will not get into an infinite loop
        assert max_retry == 0
        raise exc

    def get_initial_system_state(self):
        """ Get initial SystemState as would Dispatcher run it inside run() method """
        from selinon.system_state import SystemState
        return SystemState(id(self), self.flow_name, node_args=self.node_args, retry=None, state=None,
                           parent=self.parent, selective=self.selective)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Base class for Cache interface."""

import abc


class Cache(metaclass=abc.ABCMeta):
    """Base class for Cache classes."""

    @abc.abstractmethod
    def add(self, item_id, item, task_name=None, flow_name=None):
        """Add item to cache.

        :param item_id: item id under which item should be referenced
        :param item: item itself
        :param task_name: name of task that result should/shouldn't be cached, unused when caching Celery's AsyncResult
        :param flow_name: name of flow in which task was executed, unused when caching Celery's AsyncResult
        """

    @abc.abstractmethod
    def get(self, item_id, task_name=None, flow_name=None):
        """Get item from cache.

        :param item_id: item id under which the item is stored
        :param task_name: name of task that result should/shouldn't be cached, unused when caching Celery's AsyncResult
        :param flow_name: name of flow in which task was executed, unused when caching Celery's AsyncResult
        :return: item itself
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import logging


def get_task_logger(name):
    """Wrapper above Celery logging facilities"""
    return logging.getLogger(name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""A pool that carries all database connections for workers."""

import traceback

from .config import Config
from .errors import CacheMissError
from .errors import StorageError
from .errors import UnknownStorageError
from .lock_pool import LockPool
from .trace import Trace


class StoragePool:
    """A pool that carries all database connections for workers."""

    _storage_pool_locks = LockPool()

    def __init__(self, id_mapping, flow_name):
        """Initialize storage pool instance based on the current context.

        :param id_mapping: mapping tasks and their ids
        :param flow_name: name of flow for which StoragePool context is created
        """
        self._id_mapping = id_mapping or {}
        self._flow_name = flow_name

    @classmethod
    def get_storage_name_by_task_name(cls, task_name, graceful=False):
        """Get name of storage that was assigned to the given task.

        :param task_name: name of a task
        :param graceful: return None instead of raising an exception
        :return: storage name for task
        """
        storage = Config.task2storage_mapping.get(task_name)

        if storage is None and not graceful:
            raise UnknownStorageError("No storage for task '%s' defined" % task_name)

        return storage

    @classmethod
    def get_storage_by_task_name(cls, task_name):
        """Get storage instance that was assigned to the given task.

        :param task_name: task's name for which storage should be get
        :rtype: DataStorage
        """
        storage_name = cls.get_storage_name_by_task_name(task_name, graceful=True)

        if storage_name:
            return cls.get_connected_storage(storage_name)

        return None

    @classmethod
    def get_connected_storage(cls, storage_name):
        """Retrieve connected storage based by its name stated in configuration.

        :param storage_name: name of storage
        :return: connected storage
        """
        # if this raises UnknownStorageError exception it means that the flow was not configured properly, should
        # be handled during Python code generation.
        storage = Config.storage_mapping[storage_name]

        if not storage.is_connected():
            with cls._storage_pool_locks.get_lock(storage):
                if not storage.is_connected():
                    Trace.log(Trace.STORAGE_CONNECT, {'storage_name': storage_name})
                    storage.connect()

        return storage

    def get(self, task_name):
        """Retrieve data for task based on mapping for the current context.

        :param task_name: task's name that we are retrieving data for
        :return: task's result for the current context
        """
        return self.retrieve(self._flow_name, task_name, self._id_mapping[task_name])

    @classmethod
    def retrieve(cls, flow_name, task_name, task_id):
        """Retrieve task's result from database which was configured to be used for desired task.

        :param flow_name: flow in which the retrieval is taking place
        :param task_name: name of task for which result should be retrieved
        :param task_id: task ID to uniquely identify task results
        :return: task's result
        """
        storage = cls.get_storage_by_task_name(task_name)
        storage_task_name = Config.storage_task_name[task_name]
        storage_name = cls.get_storage_name_by_task_name(task_name)
        trace_msg = {
            'task_name': task_name,
            'storage_task_name': storage_task_name,
            'storage_name': storage_name,
            'flow_name': flow_name,
            'task_id': task_id
        }

        with cls._storage_pool_locks.get_lock(storage):
            cache = Config.storage2storage_cache[storage_name]

            result = None
            result_retrieved = False

            # Actually it is OK if there are some issues with task result cache - if there is some issue, just
            # report it in the tracing mechanism so users are aware of it and try to talk directly to storage
            # instead.
            Trace.log(Trace.TASK_RESULT_CACHE_GET, trace_msg)
            try:
                result = cache.get(task_id, task_name=storage_task_name, flow_name=flow_name)
                result_retrieved = True
            except CacheMissError:
                Trace.log(Trace.TASK_RESULT_CACHE_MISS, trace_msg, what=traceback.format_exc())
            except Exception:  # pylint: disable=broad-except
                Trace.log(Trace.TASK_RESULT_CACHE_ISSUE, trace_msg, what=traceback.format_exc())
            else:
                Trace.log(Trace.TASK_RESULT_CACHE_HIT, trace_msg)

            if not result_retrieved:
                Trace.log(Trace.STORAGE_RETRIEVE, trace_msg)
                try:
                    result = storage.retrieve(flow_name, task_name, task_id)
                except Exception as exc:
                    error_msg = "Failed to retrieve result from storage after the result was not found in cache"
                    Trace.log(Trace.STORAGE_ISSUE, trace_msg, what=traceback.format_exc())
                    raise StorageError(error_msg) from exc

            Trace.log(Trace.TASK_RESULT_CACHE_ADD, trace_msg)
            try:
                cache.add(task_id, result)
            except Exception:  # pylint: disable=broad-except
                Trace.log(Trace.TASK_RESULT_CACHE_ISSUE, trace_msg, what=traceback.format_exc())

            return result

    @classmethod
    def delete(cls, flow_name, task_name, task_id):
        """Delete task's result from database which was configured to be used for desired task.

        :param flow_name: flow in which the retrieval is taking place
        :param task_name: name of task for which result should be deleted
        :param task_id: task ID to uniquely identify task results
        """
        storage = cls.get_storage_by_task_name(task_name)
        storage_task_name = Config.storage_task_name[task_name]
        storage_name = cls.get_storage_name_by_task_name(task_name)
        trace_msg = {
            'task_name': task_name,
            'storage_task_name': storage_task_name,
            'storage_name': storage_name,
            'flow_name': flow_name,
            'task_id': task_id
        }

        with cls._storage_pool_locks.get_lock(storage):
            Trace.log(Trace.STORAGE_DELETE, trace_msg)
            try:
                storage.delete(flow_name, task_name, task_id)
            except Exception as exc:
                error_msg = "Failed to delete result from storage"
                Trace.log(Trace.STORAGE_ISSUE, trace_msg, what=traceback.format_exc())
                raise StorageError(error_msg) from exc
            Trace.log(Trace.STORAGE_DELETED, trace_msg)

        return

    @classmethod
    def set(cls, node_args, flow_name, task_name, task_id, result):
        # pylint: disable=too-many-arguments
        """Store result for task.

        :param node_args: arguments that were passed to the node
        :param flow_name: flow in which task was run
        :param task_name: task that computed result
        :param task_id: task id that computed result
        :param result: result that should be stored
        :return: result ID - a unique ID which can be used to reference task results
        """
        storage = cls.get_storage_by_task_name(task_name)
        storage_task_name = Config.storage_task_name[task_name]

        record_id = storage.store(node_args, flow_name, storage_task_name, task_id, result)
        Trace.log(Trace.STORAGE_STORE, {
            'flow_name': flow_name,
            'node_args': node_args,
            'task_name': task_name,
            'storage_task_name': storage_task_name,
            'task_id': task_id,
            'storage_name': Config.task2storage_mapping[task_name],
            'record_id': record_id
        })
        return record_id

    @classmethod
    def set_error(cls, node_args, flow_name, task_name, task_id, exc_info):
        # pylint: disable=too-many-arguments
        """Store error information for task failure.

        :param node_args: arguments that were passed to the node
        :param flow_name: flow in which task was run
        :param task_name: task that computed result
        :param task_id: task id that computed result
        :param exc_info: information about exception - tuple (type, value, traceback) as returned by sys.exc_info()
        :return: true if error was stored in database - DataStorage.store_error() was called
        """
        storage = cls.get_storage_by_task_name(task_name)
        storage_task_name = Config.storage_task_name[task_name]

        try:
            record_id = storage.store_error(node_args, flow_name, storage_task_name, task_id, exc_info)
        except NotImplementedError:
            return False

        # TODO: move conversion to string to enhanced JSON handler and rather pass objects in Trace.log()
        Trace.log(Trace.STORAGE_STORE_ERROR, {
            'flow_name': flow_name,
            'node_args': node_args,
            'task_name': task_name,
            'storage_task_name': storage_task_name,
            'task_id': task_id,
            'storage_name': Config.task2storage_mapping[task_name],
            'error_type': str(exc_info[0]),
            'error_value': str(exc_info[1]),
            'error_traceback': "".join(traceback.format_tb(exc_info[2])),
            'record_id': record_id
        })

        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################


class states:
    """
    Task states as defined in Celery (only name conforms)
    """
    PENDING = 1
    RECEIVED = 2
    STARTED = 3
    SUCCESS = 4
    FAILURE = 5
    REVOKED = 6
    REJECTED = 7
    RETRY = 8
    IGNORED = 9



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""A raw Celery task that is responsible for running SelinonTask."""

import json
import sys
import traceback

import jsonschema

from .celery import Task
from .config import Config
from .errors import FatalTaskError
from .errors import Retry
from .storage_pool import StoragePool
from .trace import Trace  # Ignore PyImportSortBear


class SelinonTaskEnvelope(Task):
    """A Celery task that is responsible for running user defined tasks from flow."""

    # Celery configuration
    ignore_result = False
    acks_late = True
    track_started = True
    max_retries = None
    name = "selinon.SelinonTaskEnvelope"

    @classmethod
    def validate_result(cls, task_name, result):
        """Validate result of the task for the given schema, if fails an Exception is raised.

        :param task_name: name of task
        :param result: result of task
        """
        schema_path = Config.output_schemas.get(task_name)
        if schema_path:
            with open(schema_path, "r") as input_file:
                schema = json.load(input_file)

            jsonschema.validate(result, schema)

    def selinon_retry(self, task_name, flow_name, parent, node_args, retry_countdown, retried_count,
                      dispatcher_id, user_retry=False):
        # pylint: disable=too-many-arguments
        """Retry on Celery level.

        :param task_name: name of the task to be retried
        :param flow_name: name of in which the task was run
        :param parent: dict of parent tasks
        :param node_args: arguments for task
        :param retry_countdown: countdown for retry
        :param retried_count: number of retries already done with this task
        :param dispatcher_id: ID id of dispatcher that is handling flow that run this task
        :param user_retry: True if retry was forced from the user
        """
        max_retry = Config.max_retry.get(task_name, 0)
        kwargs = {
            'task_name': task_name,
            'flow_name': flow_name,
            'parent': parent,
            'node_args': node_args,
            'dispatcher_id': dispatcher_id,
            'retried_count': retried_count
        }

        Trace.log(Trace.TASK_RETRY, {'flow_name': flow_name,
                                     'task_name': task_name,
                                     'task_id': self.request.id,
                                     'parent': parent,
                                     'node_args': node_args,
                                     'what': traceback.format_exc(),
                                     'retry_countdown': retry_countdown,
                                     'retried_count': retried_count,
                                     'user_retry': user_retry,
                                     'queue': Config.task_queues[task_name],
                                     'dispatcher_id': dispatcher_id,
                                     'max_retry': max_retry})
        raise self.retry(kwargs=kwargs, countdown=retry_countdown, queue=Config.task_queues[task_name])

    def run(self, task_name, flow_name, parent, node_args, dispatcher_id, retried_count=None):
        # pylint: disable=arguments-differ,too-many-arguments,too-many-locals
        """Task entry-point called by Celery.

        :param task_name: task to be run
        :param flow_name: flow in which this task run
        :param parent: dict of parent nodes
        :param node_args: node arguments within the flow
        :param dispatcher_id: dispatcher id that handles flow
        :param retried_count: number of already attempts that failed so task was retried
        :rtype: None
        """
        # we are passing args as one argument explicitly for now not to have troubles with *args and **kwargs mapping
        # since we depend on previous task and the result can be anything
        Trace.log(Trace.TASK_START, {'flow_name': flow_name,
                                     'task_name': task_name,
                                     'task_id': self.request.id,
                                     'parent': parent,
                                     'queue': Config.task_queues[task_name],
                                     'dispatcher_id': dispatcher_id,
                                     'node_args': node_args})
        try:
            task = Config.get_task_instance(
                task_name=task_name,
                flow_name=flow_name,
                parent=parent,
                task_id=self.request.id,
                dispatcher_id=dispatcher_id
            )
            result = task.run(node_args)
            self.validate_result(task_name, result)

            storage = StoragePool.get_storage_name_by_task_name(task_name, graceful=True)
            if storage and not Config.storage_readonly[task_name]:
                StoragePool.set(node_args, flow_name, task_name, self.request.id, result)
            elif result is not None:
                Trace.log(Trace.TASK_DISCARD_RESULT, {'flow_name': flow_name,
                                                      'task_name': task_name,
                                                      'task_id': self.request.id,
                                                      'parent': parent,
                                                      'node_args': node_args,
                                                      'queue': Config.task_queues[task_name],
                                                      'dispatcher_id': dispatcher_id,
                                                      'result': result})
        except Retry as retry:
            # we do not touch retried_count
            self.selinon_retry(task_name, flow_name, parent, node_args, retry.countdown, retried_count,
                               dispatcher_id, user_retry=True)
        except Exception as exc:  # pylint: disable=broad-except
            exc_info = sys.exc_info()
            max_retry = Config.max_retry.get(task_name, 0)
            retried_count = retried_count or 0

            if max_retry > retried_count and not isinstance(exc, FatalTaskError):
                retried_count += 1
                retry_countdown = Config.retry_countdown.get(task_name, 0)
                self.selinon_retry(task_name, flow_name, parent, node_args, retry_countdown, retried_count,
                                   dispatcher_id)
            else:
                Trace.log(Trace.TASK_FAILURE, {'flow_name': flow_name,
                                               'task_name': task_name,
                                               'task_id': self.request.id,
                                               'parent': parent,
                                               'node_args': node_args,
                                               'what': traceback.format_exc(),
                                               'queue': Config.task_queues[task_name],
                                               'dispatcher_id': dispatcher_id,
                                               'retried_count': retried_count})

                storage = StoragePool.get_storage_name_by_task_name(task_name, graceful=True)

                if storage and not Config.storage_readonly[task_name] \
                        and not StoragePool.set_error(node_args, flow_name, task_name, self.request.id, exc_info):
                    # TODO: move conversion to string to enhanced JSON handler and rather pass objects in Trace.log()
                    Trace.log(Trace.STORAGE_OMIT_STORE_ERROR, {
                        'flow_name': flow_name,
                        'node_args': node_args,
                        'task_name': task_name,
                        'task_id': self.request.id,
                        'error_type': str(exc_info[0]),
                        'error_value': str(exc_info[1]),
                        'error_traceback': "".join(traceback.format_tb(exc_info[2])),
                    })

                raise self.retry(max_retries=0, exc=exc)

        Trace.log(Trace.TASK_END, {'flow_name': flow_name,
                                   'task_name': task_name,
                                   'task_id': self.request.id,
                                   'parent': parent,
                                   'node_args': node_args,
                                   'queue': Config.task_queues[task_name],
                                   'dispatcher_id': dispatcher_id,
                                   'storage': StoragePool.get_storage_name_by_task_name(task_name, graceful=True)})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Errors and exceptions that can occur in Selinon code base."""

import json


class SelinonException(Exception):
    """Base Selinon exception in exception hierarchy."""


class UnsupportedCeleryError(Exception):
    """An exception raised when using unsupported Celery version."""


class FatalTaskError(SelinonException):
    """An exception that is raised by task on fatal error - task will be not retried."""


class UnknownFlowError(SelinonException):
    """Raised if there was requested or referenced flow that is not stated in the YAML configuration file."""


class UnknownStorageError(SelinonException):
    """Raised if there was requested or referenced storage that is not stated in the YAML configuration file."""


class ConfigurationError(SelinonException):
    """Raised on errors that indicate errors in the configuration files."""


class SelectiveNoPathError(SelinonException):
    """Raised when there is no path in the flow to requested node in selective task runs."""


class NoParentNodeError(SelinonException):
    """An exception raised when requested parent node (task/flow), but no such parent defined."""


class RequestError(SelinonException):
    """An error raised if there was an issue with request issued by user - usually means bad usage error."""


class UnknownError(SelinonException):
    """An error raised on unknown scenarios - possibly some bug in code."""


class Retry(SelinonException):
    """Retry task as would Celery do except you can only specify countdown for retry."""

    def __init__(self, countdown):
        """Init retry.

        :param countdown: countdown in seconds
        """
        self.countdown = countdown
        super().__init__(self, countdown)


class FlowError(SelinonException):
    """An exception that is raised once there is an error in the flow on runtime - some nodes failed."""

    def __init__(self, state):
        """Make sure flow errors capture final state of the flow.

        :param state: final flow state
        """
        super().__init__(json.dumps(state))

    @property
    def state(self):
        """Get structured flow state."""
        return json.loads(str(self))


class DispatcherRetry(SelinonException):
    """Force retry the whole flow - if flow arguments are provided, flow will continue where it was left."""

    def __init__(self, keep_state=True, adjust_retry_count=True):
        """Retry dispatcher causing flow state to be reinspected.

        :param keep_state: the current state should be kept otherwise the whole flow will start from the beginning
        :type keep_state: bool
        :param adjust_retry_count: adjust flow retry count if needed, will cause flow failure once retry count reaches 0
        :type adjust_retry_count: bool
        """
        super().__init__()
        self.keep_state = keep_state
        self.adjust_retry_count = adjust_retry_count


class StorageError(SelinonException):
    """Raised to notify about storage errors (e.g. storage goes down)."""


class MigrationNotNeeded(SelinonException):
    """Raised when a migration is requested, but config changes do not require it."""


class MigrationSkew(SelinonException):
    """Raised if worker hasn't needed migration files."""

    def __init__(self, *args, available_migration_version):
        """Instantiate and track info about migration skew.

        :param available_migration_version:
        """
        self.available_migration_version = available_migration_version
        super().__init__(*args)


class MigrationException(SelinonException):
    """Base exception for migration related exceptions."""

    TAINTED_FLOW_STRATEGY = 'UNKNOWN'

    def __init__(self, *args, migration_version, latest_migration_version, tainting_nodes=None, tainted_edge=None):
        """Instantiate base exception for migrations.

        :param args: additional arguments for base exception
        :param migration_version: the current migration version
        :param latest_migration_version: the latest migration version based on migration directory content
        :param tainting_nodes: nodes that run tainted flow
        :param tainted_edge: edges that tainted flow
        """
        self.migration_version = migration_version
        self.latest_migration_version = latest_migration_version
        self.tainting_nodes = tainting_nodes
        self.tainted_edge = tainted_edge
        super().__init__(*args)


class MigrationFlowFail(MigrationException):
    """An exception raised when a flow should fail - i.e. migration causing tainting of flow should fail flow."""

    TAINTED_FLOW_STRATEGY = 'FAIL'


class MigrationFlowRetry(MigrationException):
    """An exception raised when a flow should be retried - i.e. migration causing tainting of flow should retry flow."""

    TAINTED_FLOW_STRATEGY = 'RETRY'


class CacheMissError(SelinonException):
    """An error raised when there is requested an item from cache that is not stored in cache."""


class ConfigNotInitializedError(SelinonException):
    """An error raised when the configuration was requested, but not initialized."""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Migration of configuration files."""

from datetime import datetime
import logging
import os
import platform

import yaml

from selinon.errors import MigrationFlowFail
from selinon.errors import MigrationFlowRetry
from selinon.errors import MigrationNotNeeded
from selinon.errors import MigrationSkew
from selinon.errors import RequestError
from selinon.errors import UnknownError
from selinon.helpers import dict2json
from selinon.predicate import Predicate

from .tainted_flow_strategy import TaintedFlowStrategy

_logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Migrator:
    """Main class for performing configuration file migrations."""

    def __init__(self, migration_dir=None):
        """Initialize migrator.

        :param migration_dir: a path to directory containing migration files
        """
        self.migration_dir = migration_dir
        self.old_nodes_definition = None
        self.old_flow_definitions = {}
        self.new_nodes_definition = None
        self.new_flow_definitions = {}

    def _report_diff_flow(self):
        """Report added and removed flows."""
        new_flows = set(self.new_nodes_definition.get('flows', [])) - set(self.old_nodes_definition.get('flows', []))
        if new_flows:
            _logger.info("Newly introduced flows in your configuration: %s", ", ".join(new_flows))
            for new_flow in new_flows:
                self.new_flow_definitions.pop(new_flow)

        removed_flows = set(self.old_nodes_definition.get('flows', [])) -\
            set(self.new_nodes_definition.get('flows', []))

        if removed_flows:
            _logger.info("Removed flows from your configuration: %s", ", ".join(removed_flows))
            for removed_flow in removed_flows:
                self.old_flow_definitions.pop(removed_flow)

    def _load_flows(self, flow_files, is_old_flow=True):
        """Load flow into instance attributes.

        :param flow_files: a list of paths to flow configuration files
        :type flow_files: list
        :param is_old_flow: true if loading old configuration
        :type is_old_flow: bool
        """
        for flow_file_path in flow_files:
            with open(flow_file_path, 'r') as flow_file:
                content = yaml.safe_load(flow_file)
                for flow in content['flow-definitions']:
                    # Let's store only edges as other configuration options do not affect messages that are
                    # accepted by dispatcher
                    entry = {
                        'edges': flow['edges'],
                        'file_path': flow_file_path
                    }

                    if is_old_flow:
                        self.old_flow_definitions[flow['name']] = entry
                    else:
                        self.new_flow_definitions[flow['name']] = entry

    def _get_latest_migration_version(self):
        """Get latest migration number based on migration files present in the migration directory.

        :return: latest migration number
        """
        latest_migration_number = 0
        for file_name in os.listdir(self.migration_dir):
            file_path = os.path.join(self.migration_dir, file_name)
            if not os.path.isfile(file_path) or not file_name.endswith('.json') or file_name[0] == '.':
                _logger.debug("Skipping %r, not a file nor JSON file (or hidden file)", file_path)
                continue

            migration_number = file_name[:-len('.json')]
            try:
                migration_number = int(migration_number)
            except ValueError as exc:
                raise MigrationSkew("Unable to parse previous migrations, file name %r does not correspond "
                                    "to migration file - migration files should be named numerically"
                                    % file_path, available_migration_version=None) from exc

            latest_migration_number = max(migration_number, latest_migration_number)

        return latest_migration_number

    @staticmethod
    def _migration_file_name(migration_version):
        """Create migration file name based on migration version."""
        return str(migration_version) + ".json"

    def _get_new_migration_file_name(self):
        """Generate a new migration file name.

        :return: a name of new migration file where migrations should be stored
        :rtype: str
        """
        if not os.path.isdir(self.migration_dir):
            _logger.info("Creating migration directory %r", self.migration_dir)
            try:
                os.mkdir(self.migration_dir)
            except Exception as exc:
                raise RuntimeError("Migration directory does not exist, unable to create a new directory: %s"
                                   % str(exc))

        return self._migration_file_name(self._get_latest_migration_version() + 1)

    @staticmethod
    def _get_migration_metadata():
        """Add metadata to migration content."""
        from selinon import selinon_version
        try:
            user = os.getlogin()
        except Exception:  # pylint: disable=broad-except
            # Travis CI fails here, but let's be tolerant in other cases as well
            user = None

        return {
            'selinon_version': selinon_version,
            'host': platform.node(),
            'datetime': str(datetime.utcnow()),
            'user': user
        }

    @staticmethod
    def _is_same_migration(migration1, migration2):
        """Check if two migrations match, migration attributes that do not have effect on migration are ignored."""
        if migration1.keys() != migration2.keys():
            return False

        for flow_name in migration1.keys():
            if migration1[flow_name]['tainting_nodes'] != migration2[flow_name]['tainting_nodes']:
                return False

            if set(migration1[flow_name]['tainted_edges'].keys()) != set(migration2[flow_name]['tainted_edges'].keys()):
                return False

            if migration1[flow_name]['translation'] != migration2[flow_name]['translation']:
                return False

        return True

    def _warn_on_same_migration(self, migration):
        """Warn if the newly created migration is same as the old one."""
        # This could be optimized - we can reuse the latest migration version retrieval from call from caller.
        last_migration_version = self._get_latest_migration_version()
        if not last_migration_version:
            return

        last_migration_path = os.path.join(self.migration_dir, self._migration_file_name(last_migration_version))
        with open(last_migration_path, 'r') as last_migration_file:
            content = yaml.safe_load(last_migration_file)

        if self._is_same_migration(content['migration'], migration):
            _logger.warning("Newly created migration is same as the old configuration. "
                            "Please make sure you don't run the same migration twice!")

    def _write_migration_file(self, migration, tainted_flow_strategy, add_meta):
        """Write computed migration to migration dir."""
        new_migration_file_name = self._get_new_migration_file_name()
        new_migration_file_path = os.path.join(self.migration_dir, new_migration_file_name)

        self._warn_on_same_migration(migration)

        migration_file_content = {
            'migration': migration,
            'tainted_flow_strategy': tainted_flow_strategy.name
        }

        if add_meta:
            migration_file_content['_meta'] = self._get_migration_metadata()

        with open(new_migration_file_path, 'w') as migration_file:
            migration_file.write(dict2json(migration_file_content))

        return new_migration_file_path

    @staticmethod
    def _preprocess_edges(edges):
        """Preprocess edges before computing migrations.

        :param edges: edges from the flow YAML configuration file to be preprocessed
        """
        for idx, edge in enumerate(edges):
            for key in list(edge.keys()):
                # For debug purposes - keep condition entry so it is written to the migration and tracing module can
                # report tainted edges correctly.
                if key not in ('from', 'to', 'condition'):
                    edge.pop(key)

            edge['_idx'] = idx
            # We don't care about node order
            if not isinstance(edge['from'], (list, set)):
                edge['from'] = [edge['from']] if edge['from'] is not None else []

            if not isinstance(edge['to'], (list, set)):
                edge['to'] = [edge['to']] if edge['to'] is not None else []

            edge['from'] = set(edge['from'])
            edge['to'] = set(edge['to'])

            if 'condition' not in edge:
                edge['condition'] = Predicate.construct_default_dict()

    def _calculate_flow_migration(self, old_flow_edges, new_flow_edges):
        """Calculate migration for a flow.

        :param old_flow_edges: edges definition of old flow
        :type old_flow_edges: dict
        :param new_flow_edges: edges definition of new flow
        :param new_flow_edges: dict
        :return: tuple describing migration, tainted edges and tainting edges
        :rtype: tuple
        """
        self._preprocess_edges(old_flow_edges)
        self._preprocess_edges(new_flow_edges)

        # Let's first construct subset of edges that was not affected by user's change
        for old_edge_idx, old_edge in enumerate(old_flow_edges):
            for new_edge_idx, new_edge in enumerate(new_flow_edges):
                if new_edge.get('_old_edge_idx') is not None:
                    continue

                if old_edge['from'] == new_edge['from'] and old_edge['to'] == new_edge['to']:
                    old_edge['_new_edge_idx'] = new_edge_idx
                    new_edge['_old_edge_idx'] = old_edge_idx
                    break

        _logger.debug("All old flow edges: %s", old_flow_edges)
        _logger.debug("All new flow edges: %s", new_flow_edges)

        old_unmatched = [edge for edge in old_flow_edges if '_new_edge_idx' not in edge.keys()]
        new_unmatched = [edge for edge in new_flow_edges if '_old_edge_idx' not in edge.keys()]

        _logger.debug("Old unmatched edges: %s", old_unmatched)
        _logger.debug("New unmatched edges: %s", new_unmatched)

        # Edges that were in the old flow, but they are not in the new one - if they run, it
        # means that the flow was tainted.
        # Also be consistent with tracing mechanism - use condition_str instead of condition key.
        tainted_edges = {str(edge['_idx']): {'from': edge['from'], 'to': edge['to'], 'condition_str': edge['condition']}
                         for edge in old_unmatched}

        # Source nodes of newly added edges would cause that the newly added edge would be executed. If these nodes
        # already run, the flow is tainted.
        tainting_nodes = {str(edge['_idx']): list(edge['from']) for edge in new_unmatched if edge['from']}

        translation = {}
        for old_edge in old_unmatched:
            matched = False
            for new_edge in new_unmatched:
                if old_edge['from'] == new_edge['from']:
                    translation[str(old_edge['_idx'])] = new_edge['_idx']
                    matched = True
                    break

            if not matched:
                translation[str(old_edge['_idx'])] = None

        # Keep track of edges that were remapped to another edge
        translation.update({str(edge['_idx']): edge.get('_new_edge_idx') for edge in old_flow_edges
                            if edge['_idx'] != edge.get('_new_edge_idx')})

        return translation, tainted_edges, tainting_nodes

    def _calculate_migrations(self, tainted_flow_strategy, add_meta):
        """Calculate migration of configuration files and store output in migration directory.

        :return: a path to newly created migration file in migration directory
        """
        migrations = {}
        for flow_name in self.old_flow_definitions:
            old_flow = self.old_flow_definitions[flow_name]
            new_flow = self.new_flow_definitions[flow_name]
            migration, tainted_edges, tainting_nodes = self._calculate_flow_migration(old_flow['edges'],
                                                                                      new_flow['edges'])

            if not migrations and not tainted_edges and not tainting_nodes:
                continue

            migrations[flow_name] = {
                'translation': migration,
                'tainted_edges': tainted_edges,
                'tainting_nodes': tainting_nodes
            }

        if not migrations:
            raise MigrationNotNeeded("No flow configuration changes that would require new migration detected")

        return self._write_migration_file(migrations, tainted_flow_strategy, add_meta)

    def create_migration_file(self, old_nodes_definition_path, old_flow_definitions_path,
                              new_nodes_definition_path, new_flow_definitions_path, tainted_flow_strategy,
                              add_meta=True):
        """Generate migration of configuration files, store output in the migration directory.

        :param old_nodes_definition_path: a path to old nodes.yaml
        :type old_nodes_definition_path: str
        :param old_flow_definitions_path: a list of paths to old flow definition files
        :type old_flow_definitions_path: list
        :param new_nodes_definition_path: a path to new nodes.yaml
        :type new_nodes_definition_path: str
        :param new_flow_definitions_path: a list of paths to new flow definition files
        :type new_flow_definitions_path: list
        :param tainted_flow_strategy: flow strategy for tainted flows
        :type tainted_flow_strategy: selinon.migrator.tainted_flow_strategy.TaintedFlowStrategy
        :param add_meta: add metadata information
        :type add_meta: bool
        :return: a path to newly created migration file
        """
        _logger.info("Performing configuration files migrations, storing results in %r", self.migration_dir)

        with open(old_nodes_definition_path, 'r') as old_nodes_definition_file:
            self.old_nodes_definition = yaml.safe_load(old_nodes_definition_file)

        self._load_flows(old_flow_definitions_path, is_old_flow=True)

        with open(new_nodes_definition_path, 'r') as new_nodes_definition_file:
            self.new_nodes_definition = yaml.safe_load(new_nodes_definition_file)

        old_migration_dir = self.old_nodes_definition.get('global', {}).get('migration_dir')
        new_migration_dir = self.new_nodes_definition.get('global', {}).get('migration_dir')

        if old_migration_dir != new_migration_dir and old_migration_dir is not None:
            _logger.warning("The old configuration points to migration directory located in %r whereas new "
                            "configuration points to migration directory located in %r. Make sure you have already "
                            "moved old migrations and they are present in %r already",
                            old_migration_dir, new_migration_dir, new_migration_dir)

        if self.migration_dir is None:
            self.migration_dir = new_migration_dir
            if new_migration_dir is None:
                raise RequestError("The migration directory not explicitly provided and cannot be determined from the "
                                   "nodes.yaml configuration (%r)" % new_nodes_definition_path)

        elif self.migration_dir != new_migration_dir:
            _logger.warning("Explicitly provided migration dir %r does not match migration dir present in the "
                            "nodes.yaml file (%r): %r. Make sure your migration gets correctly "
                            "propagated on deployment.",
                            self.migration_dir, new_nodes_definition_path, new_migration_dir)

        _logger.info("Calculating config file migrations, the computed migration will be placed to %r",
                     self.migration_dir)
        self._load_flows(new_flow_definitions_path, is_old_flow=False)
        self._report_diff_flow()
        return self._calculate_migrations(tainted_flow_strategy, add_meta)

    @staticmethod
    def _do_migration(migration_spec, flow_name, state, migration_version, latest_migration_version):
        """Do single migration of message based on migration definition.

        :param migration_spec: migration specification to be used
        :param flow_name: a name of flow for which the migration is perfomed
        :param state: the current flow state
        :param migration_version: the current migration version
        :param latest_migration_version: the latest migration version available on worker
        :return:
        """
        # pylint: disable=too-many-locals
        def raise_on_tainted_state(**kwargs):
            """Raise exceptions signalizing tainted flow when needed."""
            if tainted_flow_strategy == TaintedFlowStrategy.FAIL:
                raise MigrationFlowFail("Migration requested flow to fail",
                                        migration_version=migration_version,
                                        latest_migration_version=latest_migration_version,
                                        **kwargs)

            if tainted_flow_strategy == TaintedFlowStrategy.RETRY:
                raise MigrationFlowRetry("Migration requested flow to retry",
                                         migration_version=migration_version,
                                         latest_migration_version=latest_migration_version,
                                         **kwargs)

        flow_migration = migration_spec['migration'].get(flow_name)
        if not flow_migration:
            # Nothing to do, no changes in flow in the migration
            return state, False

        waiting_edges = state.get('waiting_edges', [])
        triggered_edges = state.get('triggered_edges', [])

        if not waiting_edges and not triggered_edges:
            return state, False

        tainted_flow_strategy = TaintedFlowStrategy.get_option_by_name(migration_spec['tainted_flow_strategy'])
        tainted = False
        for idx, triggered_edge in enumerate(triggered_edges):
            if str(triggered_edge) in flow_migration['tainted_edges'].keys():
                tainted = True
                raise_on_tainted_state(tainted_edge=flow_migration['tainted_edges'][str(triggered_edge)])

            # Make sure we track triggered edges with their new transcription based on migration
            if str(triggered_edge) in flow_migration['translation']:
                state['triggered_edge'][idx] = flow_migration['translation'][str(triggered_edge)]

        # Transcript also waiting edges so they are triggered correctly based on new configuration
        for idx, waiting_edge in enumerate(waiting_edges):
            if str(waiting_edge) in flow_migration['translation']:
                state['waiting_edges'][idx] = flow_migration['translation'][str(waiting_edge)]

        # Remove edges that should be discarded (mapped to None)
        state['waiting_edges'] = list(edge for edge in state['waiting_edges'] if edge is not None)
        state['triggered_edges'] = list(edge for edge in state['triggered_edges'] if edge is not None)

        finished_and_active_nodes = set(node_name for node_name in state['finished_nodes'].keys())
        for tainting_nodes in flow_migration['tainting_nodes'].values():
            if tainting_nodes and set(tainting_nodes).issubset(finished_and_active_nodes):
                tainted = True
                raise_on_tainted_state(tainting_nodes=tainting_nodes)

        # Add edges that were added and should be triggered after the corresponding subset of active nodes finish.
        active_nodes = {node['name'] for node in state['active_nodes']}
        for edge_idx, node_names in flow_migration['tainting_nodes'].items():
            edge_idx = int(edge_idx)
            if set(node_names).issubset(active_nodes) and edge_idx not in state['waiting_edges']:
                state['waiting_edges'].append(edge_idx)

        return state, tainted

    def perform_migration(self, flow_name, state, migration_version):
        """Perform actual migration based on message received.

        :param flow_name: name of flow for which the migration is performed
        :param state: state for which migration is present
        :param migration_version: migration version that was used previously
        :return: migrated message with current migration version and information about tainting flow
        :rtype: tuple
        """
        if not self.migration_dir:
            raise UnknownError("No migration directory provided on instantiation")

        latest_migration_version = self._get_latest_migration_version()
        if migration_version is None or state is None:
            # This means that this message is consumed for the first time, adjust migration based on migration
            # version of the current worker.
            return state, latest_migration_version, False

        if migration_version == latest_migration_version:
            _logger.debug("No migrations needed, migration version is %d", migration_version)
            return state, migration_version, False

        if migration_version > latest_migration_version:
            raise MigrationSkew("Received message with a newer migration number, "
                                "message migration version: %d, latest migration number present available: %d"
                                % (migration_version, latest_migration_version),
                                available_migration_version=latest_migration_version)

        tainted = False
        current_migration_version = migration_version
        while current_migration_version != latest_migration_version:
            current_migration_version += 1
            current_migration_path = os.path.join(self.migration_dir,
                                                  self._migration_file_name(current_migration_version))

            try:
                with open(current_migration_path, 'r') as migration_file:
                    migration_spec = yaml.safe_load(migration_file)
            except FileNotFoundError as exc:
                raise MigrationSkew("Migration file %r not found, cannot perform migrations"
                                    % current_migration_path,
                                    available_migration_version=current_migration_version) from exc

            state, taints_flow = self._do_migration(migration_spec,
                                                    flow_name,
                                                    state,
                                                    current_migration_version,
                                                    latest_migration_version)
            tainted = tainted or taints_flow

        return state, current_migration_version, tainted



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Core Selinon logic for system representation, parsing and handling actions."""

import datetime
import logging
import os
import platform

import yaml

import graphviz

from .errors import ConfigurationError
from .flow import Flow
from .global_config import GlobalConfig
from .helpers import check_conf_keys
from .helpers import dict2strkwargs
from .helpers import expr2str
from .selective_run_function import SelectiveRunFunction
from .storage import Storage
from .task import Task
from .task_class import TaskClass
from .user_config import UserConfig
from .version import selinon_version

# pylint: disable=too-many-locals,too-many-nested-blocks,too-many-boolean-expressions,too-many-lines


class System:
    """The representation of the whole system."""

    _logger = logging.getLogger(__name__)

    def __init__(self, tasks=None, flows=None, storages=None, task_classes=None):
        """Initialize system info.

        :param tasks: a list of tasks available in the system
        :param flows: a list of flows available in the system
        :param storages: a list of storages available in the system
        :param task_classes: a list of classes that implement task logic (the same Python class can be represented
                             in multiple Selinon tasks)
        """
        self.flows = flows or []
        self.tasks = tasks or []
        self.storages = storages or []
        self.task_classes = task_classes or []

    def _check_name_collision(self, name):
        """All tasks and flows share name space, check for collisions.

        :param name: a node name
        :raises: ConfigurationError
        """
        if any(flow.name == name for flow in self.flows):
            raise ConfigurationError("Unable to add node with name '%s', a flow with the same name already exist"
                                     % name)
        if any(task.name == name for task in self.tasks):
            raise ConfigurationError("Unable to add node with name '%s', a task with the same name already exist"
                                     % name)

    def add_task(self, task):
        """Register a task in the system.

        :param task: a task to be registered
        :type task: selinon.selinon_task.Task
        """
        self._check_name_collision(task.name)
        self.tasks.append(task)

    def add_flow(self, flow):
        """Register a flow in the system.

        :param flow: a flow to be registered
        :type flow: flow
        """
        self._check_name_collision(flow.name)
        self.flows.append(flow)

    def add_storage(self, storage):
        """Add storage to system.

        :param storage: storage that should be added
        """
        # We need to check for name collision with tasks as well since we import them by name
        for task in self.tasks:
            if task.name == storage.name:
                raise ConfigurationError("Storage has same name as task '%s'" % storage.name)

        for stored_storage in self.storages:
            if stored_storage.name == storage:
                raise ConfigurationError("Multiple storages of the same name '{}'".format(storage.name))
        self.storages.append(storage)

    def storage_by_name(self, name, graceful=False):
        """Retrieve storage by its name.

        :param name: name of the storage
        :param graceful: if true, exception is raised if no such storage with name name is found
        :return: storage
        """
        for storage in self.storages:
            if storage.name == name:
                return storage
        if not graceful:
            raise ConfigurationError("Storage with name {} not found in the system".format(name))
        return None

    def task_by_name(self, name, graceful=False):
        """Find a task by its name.

        :param name: a task name
        :param graceful: if True, no exception is raised if task couldn't be found
        :return: task with name 'name'
        :rtype: selinon.selinon_task.Task
        :raises ConfigurationError: on non-existing task name
        """
        for task in self.tasks:
            if task.name == name:
                return task

        if not graceful:
            raise ConfigurationError("Task with name {} not found in the system".format(name))

        return None

    def flow_by_name(self, name, graceful=False):
        """Find a flow by its name.

        :param name: a flow name
        :param graceful: if True, no exception is raised if flow couldn't be found
        :return: flow with name 'name'
        :rtype: Flow
        :raises ConfigurationError: on non-existing flow name
        """
        for flow in self.flows:
            if flow.name == name:
                return flow

        if not graceful:
            raise ConfigurationError("Flow with name '{}' not found in the system".format(name))

        return None

    def node_by_name(self, name, graceful=False):
        """Find a node (flow or task) by its name.

        :param name: a node name
        :param graceful: if True, no exception is raised if node couldn't be found
        :return: flow with name 'name'
        :rtype: Node
        :raises: KeyError
        """
        node = self.task_by_name(name, graceful=True)

        if not node:
            node = self.flow_by_name(name, graceful=True)

        if not node and not graceful:
            raise ConfigurationError("Entity with name '{}' not found in the system".format(name))

        return node

    def task_queue_names(self):
        """Get information about queue names per task.

        :return: corresponding mapping from task name to task queue
        """
        ret = {}

        for task in self.tasks:
            ret[task.name] = task.queue_name

        return ret

    def dispatcher_queue_names(self):
        """Get information about queue names per dispatcher/flow.

        :return: dispatcher queue names based on flow names
        """
        ret = {}

        for flow in self.flows:
            ret[flow.name] = flow.queue_name

        return ret

    def class_of_task(self, task):
        """Return task class of a task.

        :param task: task to look task class for
        :return: TaskClass or None if a task class for task is not available
        """
        for task_class in self.task_classes:
            if task_class.task_of_class(task):
                return task_class
        return None

    def _dump_imports(self, output):
        """Dump used imports of tasks to a stream.

        :param output: a stream to write to
        """
        predicates = set()
        cache_imports = set()
        selective_run_function_imports = set()

        for flow in self.flows:
            for edge in flow.edges:
                predicates.update([p.__name__ for p in edge.predicate.predicates_used()])

            # predicates used on failure nodes
            if flow.failures:
                for predicate in flow.failures.predicates:
                    predicates.update([p.__name__ for p in predicate.predicates_used()])

            cache_imports.add((flow.cache_config.import_path, flow.cache_config.name))

        if predicates:
            output.write('from %s import %s\n' % (GlobalConfig.predicates_module, ", ".join(predicates)))

        for flow in self.flows:
            for idx, edge in enumerate(flow.edges):
                if edge.foreach:
                    output.write('from {} import {} as {}\n'.format(edge.foreach['import'],
                                                                    edge.foreach['function'],
                                                                    self._dump_foreach_function_name(flow.name, idx)))

        for task in self.tasks:
            output.write("from {} import {} as {}\n".format(task.import_path, task.class_name, task.name))
            f = task.selective_run_function  # pylint: disable=invalid-name
            selective_run_function_imports.add((f.import_path, f.name))

        for storage in self.storages:
            output.write("from {} import {}\n".format(storage.import_path, storage.class_name))
            cache_imports.add((flow.cache_config.import_path, flow.cache_config.name))

        for import_path, cache_name in cache_imports:
            output.write("from {} import {}\n".format(import_path, cache_name))

        for import_path, function_name in selective_run_function_imports:
            output.write("from {} import {} as {}\n".format(
                import_path,
                function_name,
                SelectiveRunFunction.construct_import_name(function_name, import_path)
            ))

        # we need partial for strategy function and for using storage as trace destination
        output.write("\nimport functools\n")
        # we need datetime for timedelta in throttling
        output.write("\nimport datetime\n")

    def _dump_output_schemas(self, output):
        """Dump output schema mapping to a stream.

        :param output: a stream to write to
        """
        output.write('output_schemas = {')
        printed = False
        for task in self.tasks:
            if task.output_schema:
                if printed:
                    output.write(",")
                output.write("\n    '%s': '%s'" % (task.name, task.output_schema))
                printed = True
        output.write('\n}\n\n')

    def _dump_flow_flags(self, stream):
        """Dump various flow flags.

        :param stream: a stream to write to
        """
        self._dump_dict(stream,
                        'node_args_from_first',
                        {f.name: f.node_args_from_first for f in self.flows})

        self._dump_dict(stream,
                        'propagate_node_args',
                        {f.name: f.propagate_node_args for f in self.flows})

        self._dump_dict(stream,
                        'propagate_parent',
                        {f.name: f.propagate_parent for f in self.flows})

        self._dump_dict(stream,
                        'propagate_parent_failures',
                        {f.name: f.propagate_parent_failures for f in self.flows})

        self._dump_dict(stream,
                        'propagate_finished',
                        {f.name: f.propagate_finished for f in self.flows})

        self._dump_dict(stream,
                        'propagate_compound_finished',
                        {f.name: f.propagate_compound_finished for f in self.flows})

        self._dump_dict(stream,
                        'propagate_failures',
                        {f.name: f.propagate_failures for f in self.flows})

        self._dump_dict(stream,
                        'propagate_compound_failures',
                        {f.name: f.propagate_compound_failures for f in self.flows})

    @staticmethod
    def _dump_dict(output, dict_name, dict_items):
        """Dump propagate_finished flag configuration to a stream.

        :param output: a stream to write to
        """
        output.write('%s = {' % dict_name)
        printed = False
        for key, value in dict_items.items():
            if printed:
                output.write(",")
            if isinstance(value, list):
                string = [item.name for item in value]
            else:
                string = str(value)
            output.write("\n    '%s': %s" % (key, string))
            printed = True
        output.write('\n}\n\n')

    def _dump_task_classes(self, output):
        """Dump mapping from task name to task class.

        :param output: a stream to write to
        """
        output.write('task_classes = {')
        printed = False
        for task in self.tasks:
            if printed:
                output.write(',')
            output.write("\n    '%s': %s" % (task.name, task.name))
            printed = True
        output.write('\n}\n\n')

    def _dump_storage_task_names(self, output):
        """Dump mapping for task name in storage (task name alias for storage).

        :param output: a stream to write to
        """
        output.write('storage_task_name = {')
        printed = False
        for task in self.tasks:
            if printed:
                output.write(',')
            output.write("\n    '%s': '%s'" % (task.name, task.storage_task_name))
            printed = True
        output.write('\n}\n\n')

    def _dump_queues(self, output):
        """Dump queues for tasks and dispatcher.

        :param output: a stream to write to
        """
        self._dump_dict(output, 'task_queues', {f.name: "'%s'" % f.queue_name for f in self.tasks})
        self._dump_dict(output, 'dispatcher_queues', {f.name: "'%s'" % f.queue_name for f in self.flows})

    def _dump_storage2instance_mapping(self, output):
        """Dump storage name to instance mapping to a stream.

        :param output: a stream to write to
        """
        storage_var_names = []
        for storage in self.storages:
            output.write("%s = %s" % (storage.var_name, storage.class_name))
            if storage.configuration and isinstance(storage.configuration, dict):
                output.write("(%s)\n" % dict2strkwargs(storage.configuration))
            elif storage.configuration:
                output.write("(%s)\n" % expr2str(storage.configuration))
            else:
                output.write("()\n")

            storage_var_names.append((storage.name, storage.var_name,))

        output.write('storage2instance_mapping = {\n')
        printed = False
        for storage_var_name in storage_var_names:
            if printed:
                output.write(",\n")
            output.write("    '%s': %s" % (storage_var_name[0], storage_var_name[1]))
            printed = True

        output.write("\n}\n\n")

    def _dump_task2storage_mapping(self, output):
        """Dump task name to storage name mapping to a stream.

        :param output: a stream to write to
        """
        output.write('task2storage_mapping = {\n')
        printed = False
        for task in self.tasks:
            if printed:
                output.write(",\n")
            storage_name = ("'%s'" % task.storage.name) if task.storage else str(None)
            output.write("    '%s': %s" % (task.name, storage_name))
            printed = True

        output.write("\n}\n\n")

    def _dump_storage_conf(self, output):
        """Dump storage configuration to a stream.

        :param output: a stream to write to
        """
        self._dump_task2storage_mapping(output)
        self._dump_storage2instance_mapping(output)
        for storage in self.storages:
            cache_config = storage.cache_config
            output.write("%s = %s(%s)\n" % (cache_config.var_name, cache_config.name,
                                            dict2strkwargs(cache_config.configuration)))
        self._dump_dict(output, 'storage2storage_cache', {s.name: s.cache_config.var_name for s in self.storages})

    def _dump_async_result_cache(self, output):
        """Dump Celery AsyncResult caching configuration.

        :param output: a stream to write to
        """
        for flow in self.flows:
            cache_config = flow.cache_config
            output.write("%s = %s(%s)\n" % (cache_config.var_name, cache_config.name,
                                            dict2strkwargs(cache_config.configuration)))
        self._dump_dict(output, 'async_result_cache', {f.name: f.cache_config.var_name for f in self.flows})

    def _dump_strategy_func(self, output):
        """Dump scheduling strategy function to a stream.

        :param output: a stream to write to
        """
        def strategy_func_name(flow):  # pylint: disable=missing-docstring
            return "_strategy_func_%s" % flow.name

        def strategy_func_import_name(flow):  # pylint: disable=missing-docstring
            return "_raw_strategy_func_%s" % flow.name

        strategy_dict = {}
        for flow in self.flows:
            output.write("from %s import %s as %s\n"
                         % (flow.strategy.module, flow.strategy.function, strategy_func_import_name(flow)))
            output.write('%s = functools.partial(%s, %s)\n\n'
                         % (strategy_func_name(flow),
                            strategy_func_import_name(flow),
                            dict2strkwargs(flow.strategy.func_args)))
            strategy_dict[flow.name] = strategy_func_name(flow)

        output.write('\n')
        self._dump_dict(output, 'strategies', strategy_dict)

    @staticmethod
    def _dump_condition_name(flow_name, idx):
        """Create condition name for a dump.

        :param flow_name: flow name
        :type flow_name: str
        :param idx: index of condition within the flow
        :type idx: int
        :return: condition function representation
        """
        assert idx >= 0  # nosec
        return '_condition_{}_{}'.format(flow_name, idx)

    @staticmethod
    def _dump_foreach_function_name(flow_name, idx):
        """Create foreach function name for a dump.

        :param flow_name: flow name
        :type flow_name: str
        :param idx: index of condition within the flow
        :type idx: int
        :return: condition function representation
        """
        assert idx >= 0  # nosec
        return '_foreach_{}_{}'.format(flow_name, idx)

    def _dump_condition_functions(self, output):
        """Dump condition functions to a stream.

        :param output: a stream to write to
        """
        for flow in self.flows:
            for idx, edge in enumerate(flow.edges):
                output.write('def {}(db, node_args):\n'.format(edge.predicate.construct_condition_name(flow.name, idx)))
                output.write('    return {}\n\n\n'.format(edge.predicate.to_source()))

            if flow.failures:
                flow.failures.dump_all_conditions2stream(output)

    def _dump_throttling(self, output):
        """Dump throttling configuration.

        :param output: a stream to write to
        """
        self._dump_dict(output, 'throttle_tasks', {t.name: repr(t.throttling) for t in self.tasks})
        self._dump_dict(output, 'throttle_flows', {f.name: repr(f.throttling) for f in self.flows})

    def _dump_max_retry(self, output):
        """Dump max_retry configuration to a stream.

        :param output: a stream to write to
        """
        output.write('max_retry = {')
        printed = False
        for node in self.tasks + self.flows:
            if printed:
                output.write(',')
            output.write("\n    '%s': %d" % (node.name, node.max_retry))
            printed = True
        output.write('\n}\n\n')

    def _dump_retry_countdown(self, output):
        """Dump retry_countdown configuration to a stream.

        :param output: a stream to write to
        """
        output.write('retry_countdown = {')
        printed = False
        for node in self.tasks + self.flows:
            if printed:
                output.write(',')
            output.write("\n    '%s': %d" % (node.name, node.retry_countdown))
            printed = True
        output.write('\n}\n\n')

    def _dump_storage_readonly(self, output):
        """Dump storage_readonly flow to a stream.

        :param output: a stream to write to
        """
        output.write('storage_readonly = {')
        printed = False
        for task in self.tasks:
            if printed:
                output.write(',')
            output.write("\n    '%s': %s" % (task.name, task.storage_readonly))
            printed = True
        output.write('\n}\n\n')

    def _dump_selective_run_functions(self, output):
        """Dump all selective run functions.

        :param output: a stream to write to
        """
        self._dump_dict(output,
                        'selective_run_task',
                        {t.name: t.selective_run_function.get_import_name() for t in self.tasks})

    def _dump_nowait_nodes(self, output):
        """Dump nowait nodes to a stream.

        :param output: a stream to write to
        """
        output.write('nowait_nodes = {\n')
        printed = False
        for flow in self.flows:
            if printed:
                output.write(',\n')
            output.write("    '%s': %s" % (flow.name, [node.name for node in flow.nowait_nodes]))
            printed = True

        output.write('\n}\n\n')

    def _dump_eager_failures(self, output):
        """Dump eager failures to a stream.

        :param output: a stream to write to
        """
        output.write('eager_failures = {\n')
        printed = False
        for flow in self.flows:
            if printed:
                output.write(',\n')
            output.write("    '%s': %s" % (flow.name, [node.name for node in flow.eager_failures]))
            printed = True

        output.write('\n}\n\n')

    @staticmethod
    def _dump_settings(output):
        """Dump generic settings that do not belong anywhere.

        :param output: a stream to write to
        """
        if GlobalConfig.migration_dir:
            output.write('migration_dir = %r\n' % GlobalConfig.migration_dir)
        else:
            output.write('migration_dir = None\n')

    @staticmethod
    def _dump_init(output):
        """Dump init function to a stream.

        :param output: a stream to write to
        :return:
        """
        output.write('def init(config_cls):\n')
        GlobalConfig.dump_trace(output, 'config_cls', indent_count=1)
        # Mark initialization as complete for the Config instance
        output.write('    config_cls.initialized = True\n')
        # always pass in case we have nothing to init
        output.write('    return\n')
        output.write('\n')

    def _dump_edge_table(self, output):
        """Dump edge definition table to a stream.

        :param output: a stream to write to
        """
        output.write('edge_table = {\n')
        for idx, flow in enumerate(self.flows):
            output.write("    '{}': [".format(flow.name))
            for idx_edge, edge in enumerate(flow.edges):
                if idx_edge > 0:
                    output.write(',\n')
                    output.write(' '*(len(flow.name) + 4 + 5))  # align to previous line
                output.write("{'from': %s" % str([node.name for node in edge.nodes_from]))
                output.write(", 'to': %s" % str([node.name for node in edge.nodes_to]))
                output.write(", 'condition': %s" % edge.predicate.construct_condition_name(flow.name, idx_edge))
                output.write(", 'condition_str': '%s'" % str(edge.predicate).replace('\'', '\\\''))
                if edge.foreach:
                    output.write(", 'foreach': %s" % self._dump_foreach_function_name(flow.name, idx_edge))
                    output.write(", 'foreach_str': '%s'" % edge.foreach_str())
                    output.write(", 'foreach_propagate_result': %s" % edge.foreach['propagate_result'])
                if edge.selective:
                    output.write(", 'selective': %s" % edge.selective)
                output.write("}")
            if idx + 1 < len(self.flows):
                output.write('],\n')
            else:
                output.write(']\n')
        output.write('}\n\n')

    def dump2stream(self, stream):
        """Perform system dump to a Python source code to an output stream.

        :param stream: an output stream to write to
        """
        # pylint: disable=too-many-statements
        stream.write('#!/usr/bin/env python3\n')
        stream.write('# auto-generated using Selinon v{} on {} at {}\n\n'.format(selinon_version,
                                                                                 platform.node(),
                                                                                 str(datetime.datetime.utcnow())))
        self._dump_imports(stream)
        self._dump_strategy_func(stream)
        self._dump_task_classes(stream)
        self._dump_storage_task_names(stream)
        self._dump_queues(stream)
        self._dump_storage_conf(stream)
        self._dump_async_result_cache(stream)
        self._dump_output_schemas(stream)
        self._dump_flow_flags(stream)
        self._dump_throttling(stream)
        self._dump_max_retry(stream)
        self._dump_retry_countdown(stream)
        self._dump_storage_readonly(stream)
        self._dump_selective_run_functions(stream)
        self._dump_nowait_nodes(stream)
        self._dump_eager_failures(stream)
        self._dump_init(stream)
        self._dump_condition_functions(stream)

        for flow in self.flows:
            if flow.failures:
                flow.failures.dump2stream(stream)

        stream.write('failures = {')
        printed = False
        for flow in self.flows:
            if flow.failures:
                if printed:
                    stream.write(",")
                printed = True
                stream.write("\n    '%s': %s" % (flow.name, flow.failures.starting_nodes_name(flow.name)))
        stream.write('\n}\n\n')

        self._dump_selective_run_functions(stream)
        self._dump_nowait_nodes(stream)
        self._dump_settings(stream)
        self._dump_init(stream)
        self._dump_condition_functions(stream)
        self._dump_edge_table(stream)

    def dump2file(self, output_file):
        """Perform system dump to a Python source code.

        :param output_file: an output file to write to
        """
        self._logger.debug("Performing system dump to '%s'", output_file)
        with open(output_file, 'w') as stream:
            self.dump2stream(stream)

    @staticmethod
    def _plot_connection(graph, node, condition_node, storage_connections, direction_to_condition=True, fallback=False):
        """Plot node connection to graph.

        :param graph: graph to plot to
        :param node: node to plot
        :param condition_node: condition node to connect to
        :param storage_connections: storages that were plotted with their connections (not to plot duplicit edges)
        :param direction_to_condition: direction of the connection
        :param fallback: True if plotting a fallback edge
        """
        # pylint: disable=too-many-arguments
        edge_style = UserConfig().style_fallback_edge() if fallback else UserConfig().style_edge()

        if node.is_flow():
            graph.node(name=node.name, _attributes=UserConfig().style_flow())
        else:
            graph.node(name=node.name)
            if node.storage:
                graph.node(name=node.storage.name, _attributes=UserConfig().style_storage())
                if (node.name, node.storage.name) not in storage_connections:
                    graph.edge(node.name, node.storage.name, _attributes=UserConfig().style_store_edge())
                    storage_connections.append((node.name, node.storage.name,))

        if direction_to_condition:
            graph.edge(node.name, condition_node, _attributes=edge_style)
        else:
            graph.edge(condition_node, node.name, _attributes=edge_style)

    def _plot_graph_failures(self, graph, flow, storage_connections):
        """Plot failures to existing graph.

        :param graph: graph to plot to
        :param flow: flow which failures should be plotted
        :param storage_connections: storage connections that were already defined
        """
        for idx, failure in enumerate(flow.failures.raw_definition):
            condition = flow.failures.predicates[idx]
            # This is kind-of-hack-ish as we do not want to traverse all failure nodes, so construct some unique name
            # for this particular failure condition to ensure we do not overwrite an existing one
            # See FailureNode.construct_condition_name() for a "proper" naming
            condition_node = flow.name + str(idx)
            graph.node(name=condition_node, label=str(condition), _attributes=UserConfig().style_condition())

            for node_name in failure['nodes']:
                node = self.node_by_name(node_name)
                self._plot_connection(graph, node, condition_node, storage_connections,
                                      direction_to_condition=True, fallback=True)

            if isinstance(failure['fallback'], bool):
                graph.node(name=str(id(failure['fallback'])), label=str(failure['fallback']),
                           _attributes=UserConfig().style_fallback_true())
                graph.edge(condition_node, str(id(failure['fallback'])), _attributes=UserConfig().style_fallback_edge())
            else:
                for node_name in failure['fallback']:
                    node = self.node_by_name(node_name)
                    self._plot_connection(graph, node, condition_node, storage_connections,
                                          direction_to_condition=False, fallback=True)

    def plot_graph(self, output_dir, image_format=None):  # pylint: disable=too-many-statements,too-many-branches
        """Plot system flows to graphs - each flow in a separate file.

        :param output_dir: output directory to write graphs of flows to
        :param image_format: image format, the default is svg if None
        :return: list of file names to which the graph was rendered
        :rtype: List[str]
        """
        self._logger.debug("Rendering system flows to '%s'", output_dir)
        ret = []
        image_format = image_format if image_format else 'svg'

        for flow in self.flows:
            storage_connections = []
            graph = graphviz.Digraph(format=image_format)
            graph.graph_attr.update(UserConfig().style_graph())
            graph.node_attr.update(UserConfig().style_task())
            graph.edge_attr.update(UserConfig().style_edge())

            for idx, edge in enumerate(flow.edges):
                condition_node = edge.predicate.construct_condition_name(flow.name, idx)
                if edge.foreach:
                    condition_label = "%s\n%s" % (str(edge.predicate), edge.foreach_str())
                    graph.node(name=condition_node, label=condition_label,
                               _attributes=UserConfig().style_condition_foreach())
                else:
                    graph.node(name=condition_node, label=str(edge.predicate),
                               _attributes=UserConfig().style_condition())

                for node in edge.nodes_from:
                    self._plot_connection(graph, node, condition_node, storage_connections,
                                          direction_to_condition=True)
                for node in edge.nodes_to:
                    self._plot_connection(graph, node, condition_node, storage_connections,
                                          direction_to_condition=False)

            # Plot failures as well
            if flow.failures:
                self._plot_graph_failures(graph, flow, storage_connections)

            file = os.path.join(output_dir, "%s" % flow.name)
            graph.render(filename=file, cleanup=True)
            ret.append(file)
            self._logger.info("Graph rendered to '%s.%s'", file, image_format)

        return ret

    def _post_parse_check(self):
        """Check parsed definition.

        Called once parse was done to ensure that system was correctly defined in config file.
        :raises: ValueError
        """
        self._logger.debug("Post parse check is going to be executed")
        # we want to have circular dependencies, so we need to check consistency after parsing since all flows
        # are listed (by names) in a separate definition
        for flow in self.flows:
            if not flow.edges:
                raise ConfigurationError("Empty flow: %s" % flow.name)

    def _check_propagate(self, flow):  # pylint: disable=too-many-branches
        """Check propagate configuration.

        :param flow: flow that should be checked
        :type flow: Flow
        :raises ValueError: if propagate check fails
        """
        all_source_nodes = flow.all_source_nodes()
        all_destination_nodes = flow.all_destination_nodes()
        #
        # checks on propagate_{compound_,}finished
        #
        if isinstance(flow.propagate_finished, list):
            for node in flow.propagate_finished:
                if node not in all_source_nodes:
                    raise ConfigurationError("Subflow '%s' should receive parent nodes, but there is no dependency "
                                             "in flow '%s' to which should be parent nodes propagated"
                                             % (node.name, flow.name))

                # propagate_finished set to a flow but these arguments are not passed due
                # to propagate_parent
                if node.is_flow():
                    affected_edges = [edge for edge in flow.edges if node in edge.nodes_from]
                    for affected_edge in affected_edges:
                        f = [n for n in affected_edge.nodes_to if n.is_flow()]  # pylint: disable=invalid-name
                        if len(f) == 1 and not flow.should_propagate_parent(f[0]):
                            self._logger.warning("Flow '%s' marked in propagate_finished, but calculated "
                                                 "finished nodes are not passed to sub-flow '%s' due to not "
                                                 "propagating parent, in flow '%s'",
                                                 node.name, f[0].name, flow.name)

        if isinstance(flow.propagate_compound_finished, list):
            for node in flow.propagate_compound_finished:
                if node not in all_source_nodes:
                    raise ConfigurationError("Subflow '%s' should receive parent nodes, but there is no dependency "
                                             "in flow '%s' to which should be parent nodes propagated"
                                             % (node.name, flow.name))

                # propagate_compound_finished set to a flow but these arguments are not passed due
                # to propagate_parent
                if node.is_flow():
                    affected_edges = [edge for edge in flow.edges if node in edge.nodes_from]
                    for affected_edge in affected_edges:
                        f = [n for n in affected_edge.nodes_to if n.is_flow()]  # pylint: disable=invalid-name
                        if len(f) == 1 and not flow.should_propagate_parent(f[0]):
                            self._logger.warning("Flow '%s' marked in propagate_compound_finished, but "
                                                 "calculated finished nodes are not passed to sub-flow '%s' "
                                                 "due to not propagating parent, in flow '%s'",
                                                 node.name, f[0].name, flow.name)
        #
        # checks on propagate_{compound_,}failures
        #
        # TODO: check there is set propagate_parent_failures and there is a fallback subflow to handle flow failures
        all_waiting_failure_nodes = flow.failures.all_waiting_nodes() if flow.failures else []
        if isinstance(flow.propagate_failures, list):
            for node in flow.propagate_failures:
                if node not in all_source_nodes:
                    raise ConfigurationError("Node '%s' stated in propagate_failures but this node is not started "
                                             "in flow '%s'" % (node.name, flow.name))
                if node not in all_waiting_failure_nodes:
                    raise ConfigurationError("Node '%s' stated in propagate_failures but there is no such fallback "
                                             "defined that would handle node's failure in flow '%s'"
                                             % (node.name, flow.name))

        if isinstance(flow.propagate_compound_failures, list):
            for node in flow.propagate_compound_failures:
                if node not in all_source_nodes:
                    raise ConfigurationError("Node '%s' stated in propagate_compound_failures but this node is "
                                             "not started in flow '%s'" % (node.name, flow.name))
                if node not in all_waiting_failure_nodes:
                    raise ConfigurationError("Node '%s' stated in propagate_compound_failures but there is "
                                             "no such fallback defined that would handle node's failure in flow '%s'"
                                             % (node.name, flow.name))

        if isinstance(flow.propagate_parent, list):
            for node in flow.propagate_parent:
                if node not in all_destination_nodes:
                    raise ConfigurationError("Subflow '%s' should receive parent, but there is no dependency "
                                             "in flow '%s' to which should be parent nodes propagated"
                                             % (node.name, flow.name))

        if isinstance(flow.propagate_finished, list) and isinstance(flow.propagate_compound_finished, list):
            for node in flow.propagate_finished:
                if node in flow.propagate_compound_finished:
                    raise ConfigurationError("Cannot mark node '%s' for propagate_finished and "
                                             "propagate_compound_finished at the same time in flow '%s'"
                                             % (node.name, flow.name))
        else:
            if (flow.propagate_finished is True and flow.propagate_compound_finished is True)  \
                  or (flow.propagate_finished is True and isinstance(flow.propagate_compound_finished, list)) \
                  or (isinstance(flow.propagate_finished, list) and flow.propagate_compound_finished is True):
                raise ConfigurationError("Flags propagate_compound_finished and propagate_finished are disjoint,"
                                         " please specify configuration for each node separately in flow '%s'"
                                         % flow.name)

        if isinstance(flow.propagate_failures, list) and isinstance(flow.propagate_compound_failures, list):
            for node in flow.propagate_failures:
                if node in flow.propagate_compound_failures:
                    raise ConfigurationError("Cannot mark node '%s' for propagate_failures and "
                                             "propagate_compound_failures at the same time in flow '%s'"
                                             % (node.name, flow.name))
        else:
            if (flow.propagate_failures is True and flow.propagate_compound_failures is True)  \
                  or (flow.propagate_failures is True and isinstance(flow.propagate_compound_failures, list)) \
                  or (isinstance(flow.propagate_failures, list) and flow.propagate_compound_failures is True):
                raise ConfigurationError("Flags propagate_compound_failures and propagate_failures are disjoint, "
                                         "please specify configuration for each node separately in flow '%s'"
                                         % flow.name)

    def _check(self):  # pylint: disable=too-many-statements,too-many-branches
        """Check system for consistency.

        :raises: ValueError
        """
        self._logger.info("Checking system consistency")

        for task_class in self.task_classes:
            task_ref = task_class.tasks[0]
            for task in task_class.tasks[1:]:
                if task_ref.output_schema != task.output_schema:
                    self._logger.warning("Different output schemas to a same task class: %s and %s for class '%s', "
                                         "schemas: '%s' and '%s' might differ",
                                         task_ref.name, task.name, task_class.class_name,
                                         task_ref.output_schema, task.output_schema)

                if task.max_retry != task_ref.max_retry:
                    self._logger.warning("Different max_retry assigned to a same task class: %s and %s for class '%s' "
                                         "(import '%s')",
                                         (task.name, task.max_retry), (task_ref.name, task_ref.max_retry),
                                         task_class.class_name, task_class.import_path)

        for storage in self.storages:
            if not storage.tasks:
                self._logger.warning("Storage '%s' not used in any flow", storage.name)

        all_used_nodes = set()
        # We want to check that if we depend on a node, that node is being started at least once in the flow
        # This also covers check for starting node definition
        for flow in self.flows:
            # TODO: we should make this more transparent by placing it to separate functions
            try:
                all_source_nodes = flow.all_source_nodes()
                all_destination_nodes = flow.all_destination_nodes()

                starting_edges_count = 0
                starting_nodes_count = 0

                for edge in flow.edges:
                    if not edge.nodes_from:
                        starting_edges_count += 1
                        starting_nodes_count += len(edge.nodes_to)

                        if flow.node_args_from_first:
                            if len(edge.nodes_to) > 1:
                                raise ConfigurationError("Cannot propagate node arguments from multiple "
                                                         "starting nodes")

                            if edge.nodes_to[0].is_flow():
                                raise ConfigurationError("Cannot propagate node arguments from a sub-flow")

                    node_seen = {}
                    for node_from in edge.nodes_from:
                        if not node_seen.get(node_from.name, False):
                            node_seen[node_from.name] = True
                        else:
                            raise ConfigurationError("Nodes cannot be dependent on a node of a same type mode "
                                                     "than once; node from '%s' more than once in flow '%s'"
                                                     % (node_from.name, flow.name))

                    # do not change order of checks as they depend on each other
                    edge.predicate.check()
                    edge.check()

                if starting_edges_count > 1:
                    if flow.node_args_from_first:
                        raise ConfigurationError("Cannot propagate node arguments from multiple starting nodes")

                if starting_nodes_count == 0:
                    raise ConfigurationError("No starting node found in flow '%s'" % flow.name)

                for nowait_node in flow.nowait_nodes:
                    if nowait_node in all_source_nodes:
                        raise ConfigurationError("Node '%s' marked as 'nowait' but dependency in the flow '%s' found"
                                                 % (nowait_node.name, flow.name))

                    if nowait_node not in flow.all_destination_nodes():
                        raise ConfigurationError("Node '%s' marked as 'nowait' but this node is never started in "
                                                 "flow '%s'" % (nowait_node.name, flow.name))

                for node in flow.eager_failures if isinstance(flow.eager_failures, list) else []:
                    if node not in flow.all_used_nodes():
                        raise ConfigurationError("Node %r marked as eager failure node in flow %r, but "
                                                 "this node is not present in the flow" % (node.name, flow.name))
                    if node in flow.nowait_nodes:
                        raise ConfigurationError("Node %r marked as eager failure node in flow %r, but "
                                                 "this node is also present in nowait nodes, thus it's result "
                                                 "cannot be be inspected" % (node.name, flow.name))

                    if node in flow.failures.all_waiting_nodes():
                        raise ConfigurationError("Node %r marked as eager failure node in flow %r, but "
                                                 "this node has defined a fallback, thus it cannot stop "
                                                 "flow execution eagerly" % (node.name, flow.name))

                self._check_propagate(flow)

                all_used_nodes = set(all_used_nodes) | set(all_source_nodes) | set(all_destination_nodes)
                not_started = list(set(all_source_nodes) - set(all_destination_nodes))

                error = False
                for node in not_started:
                    if node.is_task():
                        self._logger.error("Dependency in flow '%s' on node '%s', but this node is not started "
                                           "in the flow", flow.name, node.name)
                        error = True

                if error:
                    raise ConfigurationError("Dependency on not started node detected in flow '%s'" % flow.name)
            except:
                self._logger.error("Check of flow '%s' failed", flow.name)
                raise

        # we report only tasks that are not run from any flow, running a flow is based on the user
        # this check could be written more optimal by storing only tasks, but keep it this way for now
        never_started_nodes = set(self.tasks) - set(t for t in all_used_nodes if t.is_task())
        for node in never_started_nodes:
            self._logger.warning("Task '%s' (class '%s' from '%s') stated in the YAML configuration file, but "
                                 "never run in any flow", node.name, node.class_name, node.import_path)

    @classmethod
    def _setup_nodes(cls, system, nodes_definition, nodes_definition_file_name):
        """Configure nodes available in the system based on supplied configuration.

        :param system: system instance to be used
        :type system: selinon.system.System
        :param nodes_definition: a list of dictionaries holding flow configuration
        :type nodes_definition: dict
        :param nodes_definition_file_name: a name of nodes definition file (used in messages for better debugging)
        :type nodes_definition_file_name: str
        """
        def append_file_name_if_any(error_message):
            """Append file name to an error message."""
            if nodes_definition_file_name:
                error_message += ", in file %r" % nodes_definition_file_name
            return error_message

        # known top-level YAML keys for YAML config files (note flows.yml could be merged to nodes.yml)
        known_yaml_keys = ('tasks', 'flows', 'storages', 'global', 'flow-definitions')

        unknown_conf = check_conf_keys(nodes_definition, known_conf_opts=known_yaml_keys)
        if unknown_conf:
            cls._logger.warning("Unknown configuration keys in the nodes definitions, "
                                "will be skipped: %s", list(unknown_conf.keys()))

        for storage_dict in nodes_definition.get('storages', []):
            storage = Storage.from_dict(storage_dict)
            system.add_storage(storage)

        if 'global' in nodes_definition:
            GlobalConfig.from_dict(system, nodes_definition['global'])

        if 'tasks' not in nodes_definition or nodes_definition['tasks'] is None:
            raise ConfigurationError(append_file_name_if_any("No tasks defined in the system"))

        for task_dict in nodes_definition['tasks']:
            task = Task.from_dict(task_dict, system)
            task_class = system.class_of_task(task)
            if not task_class:
                task_class = TaskClass(task.class_name, task.import_path)
                system.task_classes.append(task_class)
            task_class.add_task(task)
            task.task_class = task_class
            system.add_task(task)

        if 'flows' not in nodes_definition or nodes_definition['flows'] is None:
            raise ConfigurationError(append_file_name_if_any("No flow listing defined in the system"
                                                             "in nodes definition"))

        for flow_name in nodes_definition['flows']:
            flow = Flow(flow_name)
            system.add_flow(flow)

    @classmethod
    def _setup_flows(cls, system, flow_definitions, flow_definition_file_name=None):
        """Configure Flow instances based on supplied configuration.

        :param system: system instance to be used
        :type system: selinon.system.System
        :param flow_definitions: a list of dictionaries holding flow configuration
        :type flow_definitions: list
        :param flow_definition_file_name: a name of flow definition file (used in messages for better debugging)
        :type flow_definition_file_name: str
        """
        def append_file_name_if_any(error_message):
            """Append file name to an error message."""
            if flow_definition_file_name:
                error_message += ", in file %r" % flow_definition_file_name
            return error_message

        if not isinstance(flow_definitions, (list, tuple)):
            flow_definitions = (flow_definitions,)

        for content in flow_definitions:
            flow_definitions = content.get('flow-definitions')
            if flow_definitions is None:
                error_msg = append_file_name_if_any("No flow-definitions provided in flow specification")
                raise ConfigurationError(error_msg)

            for flow_def in content['flow-definitions']:
                if 'name' not in flow_def:
                    error_msg = append_file_name_if_any("No flow name provided in the flow definition")
                    raise ConfigurationError(error_msg)

                flow = system.flow_by_name(flow_def['name'])
                try:
                    flow.parse_definition(flow_def, system)
                except:
                    error_msg = append_file_name_if_any("Failed to parse flow definition for flow %r"
                                                        % flow_def['name'])
                    cls._logger.error(error_msg)
                    raise

    @classmethod
    def from_files(cls, nodes_definition_file, flow_definition_files, no_check=False):
        """Construct System from files.

        :param nodes_definition_file: path to nodes definition file
        :type nodes_definition_file: str
        :param flow_definition_files: path to files that describe flows
        :type flow_definition_files: str
        :param no_check: True if system shouldn't be checked for consistency (recommended to check)
        :return: System instance
        :rtype: System
        """
        system = System()

        with open(nodes_definition_file, 'r') as nodes_file:
            cls._logger.debug("Parsing '%s'", nodes_definition_file)
            try:
                nodes_definition = yaml.load(nodes_file, Loader=yaml.SafeLoader)
            except Exception as exc:
                error_msg = "Bad YAML file, unable to load %r: %s" % (nodes_definition_file, str(exc))
                raise ConfigurationError(error_msg) from exc

            cls._setup_nodes(system, nodes_definition, nodes_definition_file)

        if isinstance(flow_definition_files, str):
            flow_definition_files = (flow_definition_files,)

        for flow_file in flow_definition_files:
            with open(flow_file, 'r') as flow_definition:
                cls._logger.debug("Parsing '%s'", flow_file)
                try:
                    content = yaml.load(flow_definition, Loader=yaml.SafeLoader)
                except Exception as exc:
                    error_msg = "Bad YAML file, unable to load flow from %r: %s" % (flow_file, str(exc))
                    raise ConfigurationError(error_msg) from exc

                cls._setup_flows(system, content, flow_file)

        system._post_parse_check()  # pylint: disable=protected-access
        if not no_check:
            system._check()  # pylint: disable=protected-access

        return system

    @classmethod
    def from_dict(cls, nodes_definition, flow_definitions, no_check=False):
        """Construct System from dictionaries.

        :param nodes_definition: parsed nodes definition
        :type nodes_definition: dict
        :param flow_definitions: a list of parsed flow definitions
        :type nodes_definition: list
        :param no_check: True if system shouldn't be checked for consistency (recommended to check)
        :return: System instance
        :rtype: System
        """
        system = System()

        cls._setup_nodes(system, nodes_definition, nodes_definition_file_name=None)

        for flow_def in flow_definitions:
            cls._setup_flows(system, flow_def, flow_definition_file_name=None)

        system._post_parse_check()  # pylint: disable=protected-access
        if not no_check:
            system._check()  # pylint: disable=protected-access

        return system



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Last-In-First-Out cache implementation."""

from .fifo import FIFO


class LIFO(FIFO):
    """Last-In-First-Out cache - based on FIFO implementation."""

    def _clean_cache(self):
        """Trim cache."""
        while self.current_cache_size + 1 > self.max_cache_size and self.current_cache_size > 0:
            latest = self._cache_usage.pop()
            del self._cache[latest]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Pool of all queues in the system."""

from .time_queue import TimeQueue


class QueuePool:
    """Pool of all queues in the system."""

    class _QueueWrapper:
        """Wrap a queue so we carry additional info needed for QueuePool - cyclic double linked list."""

        def __init__(self, previous_wrapper, next_wrapper, queue_name):
            """Init QueueWrapper.

            :param previous_wrapper: preceding queue wrapper in the linked list
            :param next_wrapper: next queue wrapper in the linked list
            :param queue_name: name of the queue that is wrapped
            """
            self.queue_name = queue_name
            self.queue = TimeQueue()
            self.previous = previous_wrapper or self
            self.next = next_wrapper or self

        def __next__(self):  # noqa
            return self.next

        def __repr__(self):
            """Queue representation for nice logs.

            :return: queue representation
            """
            return "%s(queue='%s')" % (self.__class__.__name__, self.queue_name)

    def __init__(self):
        """Initialize pool of queues."""
        # Queues are instantiated lazily on demand.
        self._queues = {}
        self._last_used = None
        self._queue_head = None
        self._queue_tail = None

    def _create_queue_wrapper(self, queue_name):
        """Create queue wrapper for the given queue name and register it to our queue pool (cyclic double-linked list).

        :param queue_name: queue name that the wrapper represents
        :return queue wrapper
        """
        queue_wrapper = self._QueueWrapper(self._queue_tail, self._queue_head, queue_name)

        if self._queue_tail is not None:
            self._queue_tail.next = queue_wrapper
        self._queue_tail = queue_wrapper

        if self._queue_head is None:
            self._queue_head = queue_wrapper
        self._queue_head.previous = queue_wrapper

        return queue_wrapper

    def _remove_queue_wrapper(self, queue_wrapper):
        """Remove queue wrapper from queue pool - double linked list."""
        self._queues.pop(queue_wrapper.queue_name)

        if queue_wrapper is self._queue_head and queue_wrapper is self._queue_tail:
            self._queue_head = None
            self._queue_tail = None
            self._last_used = None
            return

        queue_wrapper.previous.next = queue_wrapper.next
        queue_wrapper.next.previous = queue_wrapper.previous

        if queue_wrapper is self._queue_head:
            self._queue_head = queue_wrapper.next

        if queue_wrapper is self._queue_tail:
            self._queue_tail = queue_wrapper.previous

        if queue_wrapper is self._last_used:
            self._last_used = queue_wrapper.previous

    def get_queue(self, name):
        """Get queue wrapper by name of the queue that is wrapped, if does not exist, create one lazily.

        :param name: a name of the queue
        :return: queue wrapper for the given queue with requested name
        """
        queue_wrapper = self._queues.get(name)
        if queue_wrapper is None:
            queue_wrapper = self._create_queue_wrapper(name)
            self._queues[name] = queue_wrapper
        return queue_wrapper

    def queue_exists(self, name):
        """Check whether a queue with the given name exists.

        :param name: name of queue to be checked if exists
        :return: True if such queue exists in queue pool
        """
        return name in self._queues

    def push(self, queue_name, time, record):
        """Push record with its time to queue with name queue_name.

        :param queue_name: a queue name that should keep record
        :param time: time of record (when should be record executed)
        :param record: record itself (message with additional information such as task name and its parameters)
        """
        # TODO: add lock per queue
        queue_wrapper = self.get_queue(queue_name)
        queue_wrapper.queue.push(time, record)

        # initialize head if we haven't pushed anything to queue yet
        # so we start from the very first queue
        if self._last_used is None:
            self._last_used = queue_wrapper

    def pop(self):
        """Pop a record with the smallest time.

        :return: (time, record) tuple -  time of record and record itself (see self.push for more info)
        """
        # TODO: add lock per queue
        assert self._last_used is not None  # nosec

        exit_loop = False
        result_time, result_record = self._last_used.queue.top()
        result_queue_wrapper = self._last_used
        queue_wrapper = next(self._last_used)

        while not exit_loop:
            candidate_time, candidate_record = queue_wrapper.queue.top()

            if result_time is None or candidate_time < result_time:
                result_time, result_record = candidate_time, candidate_record
                result_queue_wrapper = queue_wrapper

            queue_wrapper = next(queue_wrapper)
            if queue_wrapper == self._last_used:
                exit_loop = True

        result_queue_wrapper.queue.pop()

        if result_queue_wrapper.queue.is_empty():
            self._remove_queue_wrapper(result_queue_wrapper)
        else:
            self._last_used = result_queue_wrapper

        return result_time, result_record

    def is_empty(self):
        """Check pool emptiness.

        :return: True if queue pool does not have any messages and no queues (as queues get deleted when empty)
        """
        return len(self._queues) == 0

    def __repr__(self):
        """Queue pool representation.

        :return: a string representing pool (with a list of active queues)
        """
        return "%s(%s)" % (self.__class__.__name__, self._queues)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Global lock pool for shared locks."""

from multiprocessing import Lock


class LockPool:  # pylint: disable=too-few-public-methods
    """Lock pool for shared locks."""

    def __init__(self):
        """Initialize lock-pool."""
        self._locks = {}
        # Instance global lock to defend operations on stored locks
        self._global_lock = Lock()

    def get_lock(self, lock_id):
        """Get lock for resource, exclusively.

        :param lock_id: lock_id for uniquely identify lock
        :return: lock, can be acquired if already taken, if new, always released
        """
        # TODO: we could make it more optimal by storing only locks that are used and release them once unused
        with self._global_lock:
            if not self._locks.get(lock_id):
                ret = Lock()
                self._locks[lock_id] = ret
            else:
                ret = self._locks[lock_id]

        return ret



#!/usr/bin/env python3
"""A wrapper around imports of Celery so Celery is not required to be installed."""

# This is a hack, make pylint happy - we know what we are doing...
# pylint: disable=unused-argument,missing-docstring,unused-import


def _raise_import_exception():
    """Raise an exception signalizing Celery not be present."""
    raise ImportError("Celery not installed, if you plan to use Selinon with Celery "
                      "install Celery using pip3 install selinon[celery]")


try:
    from celery.result import AsyncResult
except ImportError as exception:
    class AsyncResult:
        """Wrap AsyncResult so other parts of Selinon can work (e.g. executor)."""

        def __init__(self):
            raise _raise_import_exception()

try:
    from celery import Task
except ImportError:
    class Task:
        """Substitute Celery's task so we do not fail with importing - raise once usage is requested."""

        def __init__(self):
            self.request = id(self)

        @staticmethod
        def apply_async(*args, **kwargs):
            raise _raise_import_exception()

        @staticmethod
        def retry(*args, **kwargs):
            raise _raise_import_exception()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Test creation and generation of migration files.

All inputs/outputs used in these sources are present at test/data/migration/input and test/data/migration/output
respectively. Methods have "empty" implementation as name of tests distinguish file location and test decorators
specify test case expectations (failure/success).
"""

import json
import os
import pytest
import shutil
import tempfile

from selinon.errors import MigrationNotNeeded
from selinon.migrations import Migrator
from selinon.migrations import TaintedFlowStrategy

from selinon_test_case import SelinonTestCase


def migration_creation_test(tainted_flow_strategy=None):
    """A migration test (successful) execution. Files are expected to be named based on test name."""
    def decorator(test):
        def wrapper(test_instance):
            old_config_files = test_instance.get_test_config_files(test.__name__, new_config_files=False)
            new_config_files = test_instance.get_test_config_files(test.__name__, new_config_files=True)
            reference_output = test_instance.get_reference_test_output(test.__name__)

            migrator = Migrator(test_instance.migration_dir)
            migration_file_path = migrator.create_migration_file(
                old_config_files[0],
                old_config_files[1],
                new_config_files[0],
                new_config_files[1],
                tainted_flow_strategy or TaintedFlowStrategy.get_default_option()
            )

            test_instance.check_migration_match(migration_file_path, reference_output)
            test(test_instance)

        return wrapper

    return decorator


def migration_test_exception(exc, tainted_flow_strategy=None):
    """A migration test (exception) execution. Files are expected to be named based on test name."""
    def decorator(test):
        def wrapper(test_instance):
            old_config_files = test_instance.get_test_config_files(test.__name__, new_config_files=False)
            new_config_files = test_instance.get_test_config_files(test.__name__, new_config_files=True)
            migrator = Migrator(test_instance.migration_dir)

            with pytest.raises(exc):
                migrator.create_migration_file(
                    old_config_files[0],
                    old_config_files[1],
                    new_config_files[0],
                    new_config_files[1],
                    tainted_flow_strategy or TaintedFlowStrategy.get_default_option()
                )

        return wrapper

    return decorator


class TestCreateMigration(SelinonTestCase):
    """Test creation and generation of coniguration migrations."""

    migration_dir = None

    def setup_method(self):
        """Set up migration test - create a temporary migration directory."""
        self.migration_dir = tempfile.mkdtemp(prefix='migration_dir_')

    def teardown_method(self):
        """Clear all temporary files present in the migration directory after each method."""
        shutil.rmtree(self.migration_dir)

    def get_test_config_files(self, test_name, new_config_files=True):
        """Get configuration files for specific test case - path respects test method naming.

        :param test_name: name of test that is going to be executed
        :param new_config_files: True if new config files should be listed
        :return: path to nodes yaml and flow configuration files
        :rtype: tuple
        """
        # Force structure of test data
        test_dir = os.path.join(
            self.DATA_DIR,
            'migrations',
            'input',
            'new' if new_config_files else 'old',
            test_name[len('test_'):]
        )

        nodes_yaml = os.path.join(test_dir, 'nodes.yaml')
        flows_yaml = []
        flows_yaml_path = os.path.join(os.path.join(test_dir, 'flows'))
        for flow_file in os.listdir(flows_yaml_path):
            if flow_file.endswith('.yaml'):
                flows_yaml.append(os.path.join(flows_yaml_path, flow_file))

        return nodes_yaml, flows_yaml

    def get_reference_test_output(self, test_name):
        """Get output for test - path naming respects test method naming."""
        return os.path.join(
            self.DATA_DIR,
            'migrations',
            'output',
            "{test_name}.json".format(test_name=test_name[len('test_'):])
        )

    @staticmethod
    def check_migration_match(migration_file_path, reference_migration_file_path):
        """Check that two migrations match and migration metadata precedence."""
        with open(migration_file_path, 'r') as migration_file:
            migration = json.load(migration_file)

        with open(reference_migration_file_path, 'r') as reference_migration_file:
            reference_migration = json.load(reference_migration_file)

        assert 'migration' in reference_migration
        assert 'migration' in migration

        assert set(reference_migration['migration'].keys()) == set(migration['migration'].keys())

        # we check only migration, metadata may vary
        for flow_name in migration['migration'].keys():
            # required fields
            assert 'tainted_edges' in migration['migration'][flow_name], \
                "Migration check failed in flow %r" % flow_name
            json.dumps(migration['migration'][flow_name]['tainted_edges'])
            assert set(reference_migration['migration'][flow_name]['tainted_edges'].keys()) \
                   == set(migration['migration'][flow_name]['tainted_edges'].keys()), \
                "Migration check failed in flow %r" % flow_name

            assert 'tainting_nodes' in migration['migration'][flow_name], \
                "Migration check failed in flow %r" % flow_name
            assert set(reference_migration['migration'][flow_name]['tainting_nodes']) \
                   == set(migration['migration'][flow_name]['tainting_nodes']), \
                "Migration check failed in flow %r" % flow_name

            assert 'translation' in migration['migration'][flow_name], \
                "Migration check failed in flow %r" % flow_name
            assert reference_migration['migration'][flow_name]['translation']\
                   == reference_migration['migration'][flow_name]['translation'], \
                "Migration check failed in flow %r" % flow_name

        assert 'tainted_flow_strategy' in reference_migration
        assert 'tainted_flow_strategy' in migration
        assert reference_migration['tainted_flow_strategy'] == migration['tainted_flow_strategy']
        assert migration['tainted_flow_strategy'] in TaintedFlowStrategy.get_option_names()

        # check for metadata precedence
        assert '_meta' in migration, "No metadata present in the output!"
        assert 'datetime' in migration['_meta']
        assert 'host' in migration['_meta']
        assert 'selinon_version' in migration['_meta']
        assert 'user' in migration['_meta']

    @migration_test_exception(MigrationNotNeeded)
    def test_no_change(self):
        """Test that if no change is done in configuration files, an exception is raised."""

    @migration_test_exception(MigrationNotNeeded)
    def test_add_to(self):
        """Test adding source node - this doesn't require any migrations to be present."""

    @migration_test_exception(MigrationNotNeeded)
    def test_remove_to(self):
        """Test removing source node - this doesn't require any migrations to be present."""

    @migration_test_exception(MigrationNotNeeded)
    def test_add_flow(self):
        """Test adding a new flow - this doesn't require any migrations to be present."""

    @migration_creation_test(TaintedFlowStrategy.IGNORE)
    def test_add_from(self):
        """Test adding source node - edge should be discarded (mapping to None)."""

    @migration_creation_test(TaintedFlowStrategy.IGNORE)
    def test_remove_from(self):
        """Test removing source node - edge should be discarded (mapping to None)."""

    @migration_creation_test(TaintedFlowStrategy.RETRY)
    def test_remove_edge(self):
        """Test removing whole edge definition.

        Edge should be discarded (mapping to None) and all subsequent indexes shifted.
        """

    @migration_creation_test(TaintedFlowStrategy.FAIL)
    def test_add_edge(self):
        """Test adding whole edge definition.

        Edge should be discarded (mapping to None) and all subsequent indexes shifted.
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Least-Recently-Used cache implementation."""

from selinon import Cache
from selinon.errors import CacheMissError


class _Record:
    """Record that is used in a double-linked list in order to track usage."""

    def __init__(self, item_id, item):  # noqa
        self.item_id = item_id
        self.item = item
        self.previous = None
        self.next = None

    def __repr__(self):  # noqa
        return "<%s>" % self.item_id


class LRU(Cache):
    """Least-Recently-Used cache."""

    def __init__(self, max_cache_size):
        """Initialize cache.

        :param max_cache_size: maximum number of items stored in the cache
        """
        # let's allow zero size
        assert max_cache_size >= 0  # nosec

        self.max_cache_size = max_cache_size
        self._cache = {}

        self._record_head = None
        self._record_tail = None
        self.current_cache_size = 0

    def __repr__(self):
        """Cache representation for logs/debug.

        :return: string representation of cache
        """
        records = []

        record = self._record_head
        while record:
            records.append(record.item_id)
            record = record.previous

        return "%s(%s)" % (self.__class__.__name__, records)

    def _add_record(self, record):
        """Add record to cache, record shouldn't be present in the cache.

        :param record: record to add to cache
        """
        self._cache[record.item_id] = record
        self.current_cache_size += 1

        if not self._record_head:
            self._record_head = record

        if self._record_tail:
            record.next = self._record_tail
            self._record_tail.previous = record

        record.next = self._record_tail
        self._record_tail = record

    def _remove_record(self, record):
        """Remove record from cache, record should be present in the cache.

        :param record: record to be deleted
        """
        del self._cache[record.item_id]
        self.current_cache_size -= 1

        if record.next:
            record.next.previous = record.previous

        if record.previous:
            record.previous.next = record.next

        if record == self._record_tail:
            self._record_tail = record.next

        if record == self._record_head:
            self._record_head = record.previous

        record.next = None
        record.previous = None

    def _clean_cache(self):
        """Trim cache."""
        while self.current_cache_size + 1 > self.max_cache_size and self.current_cache_size > 0:
            self._remove_record(self._record_head)

    def add(self, item_id, item, task_name=None, flow_name=None):
        """Add item to cache.

        :param item_id: item id under which item should be referenced
        :param item: item itself
        :param task_name: name of task that result should/shouldn't be cached, unused when caching Celery's AsyncResult
        :param flow_name: name of flow in which task was executed, unused when caching Celery's AsyncResult
        """
        if item_id in self._cache:
            # we mark usage only in get()
            return

        self._clean_cache()

        if self.max_cache_size > 0:
            record = _Record(item_id, item)
            self._add_record(record)

    def get(self, item_id, task_name=None, flow_name=None):
        """Get item from cache.

        :param item_id: item id under which the item is stored
        :param task_name: name of task that result should/shouldn't be cached, unused when caching Celery's AsyncResult
        :param flow_name: name of flow in which task was executed, unused when caching Celery's AsyncResult
        :return: item itself
        """
        record = self._cache.get(item_id)

        # this if is safe as we store tuple - we handle even None results
        if not record:
            raise CacheMissError()

        # mark record usage
        self._remove_record(record)
        self._add_record(record)

        return record.item



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

from selinon import SelinonTask


class Task1(SelinonTask):
    def run(self, node_args):
        pass


class task2(SelinonTask):
    # class name left lowercase intentionally
    def run(self, node_args):
        pass


class task3(SelinonTask):
    # class name left lowercase intentionally
    def run(self, node_args):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Selinon Dispatcher task implementation."""

import traceback

from .celery import Task
from .config import Config
from .errors import DispatcherRetry
from .errors import FlowError
from .errors import MigrationException
from .errors import MigrationFlowFail
from .errors import MigrationFlowRetry
from .errors import MigrationSkew
from .migrations import Migrator
from .system_state import SystemState
from .trace import Trace


class Dispatcher(Task):
    """Selinon Dispatcher worker implementation."""

    # Celery configuration
    ignore_result = False
    acks_late = True
    track_started = True
    max_retries = None
    name = "selinon.Dispatcher"

    def flow_failure(self, state):
        """Mark the whole flow as failed ignoring retry configuration.

        :param state: flow state that should be captured
        :raises celery.exceptions.Retry: Celery's retry exception, always
        """
        reported_state = {
            'finished_nodes': (state or {}).get('finished_nodes', {}),
            'failed_nodes': (state or {}).get('failed_nodes', {}),
            'active_nodes': (state or {}).get('active_nodes', [])
        }
        exc = FlowError(reported_state)
        raise self.retry(max_retries=0, exc=exc)

    def selinon_retry(self, flow_info, adjust_retried_count=True, keep_state=True):
        """Retry whole flow on failure if configured so, forget any progress done so far.

        :param flow_info: a dictionary holding all the information relevant to flow (dispatcher arguments)
        :type flow_info: dict
        :param adjust_retried_count: if true, retried count will be adjusted, could cause flow failure
        :type adjust_retried_count: bool
        :param keep_state: keep or discard the current progress of the flow
        :type keep_state: bool
        :raises celery.Retry: always
        """
        new_retried_count = flow_info.get('retried_count')
        if adjust_retried_count:
            new_retried_count += 1

        if not keep_state:
            flow_info['state'] = None
            flow_info['retried_count'] = None
            flow_info['retry'] = None

        kwargs = {
            'flow_name': flow_info['flow_name'],
            'node_args': flow_info['node_args'],
            'parent': flow_info['parent'],
            'retried_count': new_retried_count,
            'selective': flow_info['selective'],
            'retry': flow_info['retry'],
            'state': flow_info['state']
        }
        countdown = Config.retry_countdown.get(flow_info['flow_name'], 0)
        max_retry = Config.max_retry.get(flow_info['flow_name'], 0)

        if new_retried_count > max_retry:
            # Force max_retries to 0 so we are not scheduled and marked as FAILED
            raise self.flow_failure(flow_info['state'])

        # We will force max retries to None so we are always retried by Celery
        queue = Config.dispatcher_queues[flow_info['flow_name']]
        # TODO: add exception here as well
        Trace.log(Trace.FLOW_RETRY, kwargs, countdown=countdown, queue=queue)
        raise self.retry(
            kwargs=kwargs,
            max_retries=None,
            countdown=countdown,
            queue=queue
        )

    def migrate_message(self, flow_info):
        """Perform migration of state first before proceeding.

        :param flow_info: information about the current flow
        """
        if Config.migration_dir:
            migrator = Migrator(Config.migration_dir)

            try:
                state, current_migration_version, tainted = migrator.perform_migration(
                    flow_info['flow_name'],
                    flow_info['state'],
                    flow_info['migration_version']
                )
            except MigrationException as exc:
                Trace.log(
                    Trace.MIGRATION_TAINTED_FLOW,
                    flow_info,
                    migration_version=exc.migration_version,
                    latest_migration_version=exc.latest_migration_version,
                    tainting_nodes=exc.tainting_nodes,
                    tainted_edge=exc.tainted_edge,
                    tainted_flow_strategy=exc.TAINTED_FLOW_STRATEGY
                )

                if isinstance(exc, MigrationFlowRetry):
                    raise self.selinon_retry(flow_info, adjust_retried_count=False, keep_state=False)

                if isinstance(exc, MigrationFlowFail):
                    raise self.flow_failure(flow_info['state'])

                raise self.flow_failure(flow_info['state'])
            except MigrationSkew as exc:
                Trace.log(Trace.MIGRATION_SKEW, flow_info, available_migration_version=exc.available_migration_version)
                raise self.selinon_retry(flow_info, adjust_retried_count=False)
            except Exception:
                # If there is anything wrong with migrations, give it a try to be fixed, retry.
                Trace.log(Trace.MIGRATION_ERROR, flow_info, what=traceback.format_exc())
                raise self.selinon_retry(flow_info, adjust_retried_count=False, keep_state=True)

            # Report success of migration
            if current_migration_version != flow_info['migration_version']:
                Trace.log(
                    Trace.MIGRATION,
                    flow_info,
                    new_migration_version=current_migration_version,
                    old_migration_version=flow_info['migration_version'],
                    tainted=tainted
                )
                # Update flow info so we are up2date with migration performed
                flow_info['migration_version'] = current_migration_version
                flow_info['state'] = state

        elif flow_info['migration_version']:
            # Keep retrying so hopefully some node in the cluster is able to proceed with this message.
            Trace.log(Trace.MIGRATION_SKEW, flow_info, available_migration_version=None)
            raise self.selinon_retry(flow_info, adjust_retried_count=False)

    def run(self, flow_name, node_args=None, parent=None, retried_count=None, retry=None,
            state=None, selective=False, migration_version=None):
        # pylint: disable=too-many-arguments,arguments-differ,too-many-locals
        """Dispatcher entry-point - run each time a dispatcher is scheduled.

        :param flow_name: name of the flow
        :param parent: flow parent nodes
        :param node_args: arguments for workers
        :param retried_count: number of Selinon retries done (not Celery retries)
        :param retry: last retry countdown
        :param state: the current system state
        :param selective: selective flow information if run in selective flow
        :param migration_version: migration version that was used for the flow
        :raises: FlowError
        """
        retried_count = retried_count or 0
        flow_info = {
            'flow_name': flow_name,
            'dispatcher_id': self.request.id,
            'node_args': node_args,
            'retry': retry,
            'queue': Config.dispatcher_queues[flow_name],
            'state': state,
            'selective': selective,
            'retried_count': retried_count,
            'parent': parent,
            'migration_version': migration_version or 0
        }

        Trace.log(Trace.DISPATCHER_WAKEUP, flow_info)

        # Perform migrations at first place
        self.migrate_message(flow_info)

        try:
            system_state = SystemState(self.request.id, flow_name, node_args, retry, state, parent, selective)
            retry = system_state.update()
        except FlowError as exc:
            max_retry = Config.max_retry.get(flow_name, 0)
            Trace.log(Trace.FLOW_FAILURE, flow_info, state=exc.state, will_retry=retried_count < max_retry)
            raise self.selinon_retry(
                flow_info=flow_info,
                adjust_retried_count=True,
                keep_state=False
            )
        except DispatcherRetry as exc:
            raise self.selinon_retry(flow_info, exc.adjust_retry_count, keep_state=exc.keep_state)
        except Exception:
            Trace.log(Trace.DISPATCHER_FAILURE, flow_info, what=traceback.format_exc())
            raise self.flow_failure(state)

        state_dict = system_state.to_dict()
        node_args = system_state.node_args

        if retry is not None and retry >= 0:
            kwargs = {
                'flow_name': flow_name,
                'node_args': node_args,
                'parent': parent,
                'retried_count': retried_count,
                'retry': retry,
                'state': state_dict,
                'selective': system_state.selective,
                'migration_version': flow_info['migration_version']
            }
            Trace.log(Trace.DISPATCHER_RETRY, flow_info, kwargs)
            raise self.retry(args=[], kwargs=kwargs, countdown=retry, queue=Config.dispatcher_queues[flow_name])

        Trace.log(Trace.FLOW_END, flow_info, state=state_dict)
        return {
            'finished_nodes': state_dict['finished_nodes'],
            # This is always {} since we have finished, but leave it here because of failure tracking.
            'failed_nodes': state_dict['failed_nodes'],
            # Always an empty array.
            'active_nodes': state_dict.get('active_nodes', [])
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Built-in predicates used as core building blocks to build predicates."""

import abc
import ast
from functools import reduce

from .errors import ConfigurationError
from .predicate import Predicate


class BuiltinPredicate(Predicate, metaclass=abc.ABCMeta):  # pylint: disable=abstract-method
    """Build in predicate abstract class."""


class NaryPredicate(BuiltinPredicate, metaclass=abc.ABCMeta):  # pylint: disable=abstract-method
    """N-ary predicate abstract class."""

    def __init__(self, children):
        """Instantiate N-ary predicate.

        :param children: children (predicates) of this predicate
        """
        super().__init__()
        self._children = children

    def _str(self, operator):
        """Create a string representation of this predicate.

        :param operator: boolean operator that should be used for concatenating child predicates
        :return: string representation (Python code)
        :rtype: str
        """
        ret = ""
        for child in self._children:
            if ret:
                ret += " %s " % operator
            ret += str(child)

        if len(self._children) > 1:
            ret = "(" + ret + ")"

        return ret

    @staticmethod
    def _create(tree, cls, nodes_from, flow, can_inspect_results):
        """Instantiate N-ary predicate class.

        :param tree: node from which should be predicate instantiated
        :type tree: List
        :param cls: class of type NaryPredicate
        :param nodes_from: nodes that are used in described edge
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :param can_inspect_results: True if predicates in the condition can query task result
        :type can_inspect_results: bool
        :return: instance of cls
        """
        if not isinstance(tree, list):
            raise ConfigurationError("Nary logical operators expect list of children")
        children = []
        for child in tree:
            children.append(Predicate.construct(child, nodes_from, flow, can_inspect_results))
        return cls(children)

    def predicates_used(self):
        """Compute predicates that are used (transitively).

        :return: used predicates by children
        """
        return reduce(lambda x, y: x + y.predicates_used(), self._children, [])

    def nodes_used(self):
        """Compute nodes that are used (transitively).

        :return: list of nodes that are used
        :rtype: List[Node]
        """
        return reduce(lambda x, y: x + y.nodes_used(), self._children, [])

    def check(self):
        """Check predicate for consistency."""
        for child in self._children:
            child.check()

    def requires_message(self):
        """Check whether any of child predicates requires results of nodes.

        :return: True if any of the children require results of parent task
        """
        return any(child.requires_message() for child in self._children)


class UnaryPredicate(BuiltinPredicate, metaclass=abc.ABCMeta):  # pylint: disable=abstract-method
    """Unary predicate abstract class."""

    def __init__(self, child):
        """Instantiate unary predicate.

        :param child: child (predicate) of this predicate
        """
        super().__init__()
        self._child = child

    @staticmethod
    def _create(tree, cls, nodes_from, flow, can_inspect_results):
        """Instantiate N-ary predicate class.

        :param tree: node from which should be predicate instantiated
        :type tree: List
        :param cls: class of type NaryPredicate
        :param nodes_from: nodes that are used in described edge
        :param flow: flow to which predicate belongs to
        :param can_inspect_results: True if predicates in the condition can query task result
        :type can_inspect_results: bool
        :return: instance of cls
        """
        if isinstance(tree, list):
            raise ConfigurationError("Unary logical operators expect one child")
        return cls(Predicate.construct(tree, nodes_from, flow, can_inspect_results))

    def predicates_used(self):
        """Compute all predicates that are used (transitively) by child/children.

        :return: used predicates by children
        """
        return self._child.predicates_used()

    def nodes_used(self):
        """Compute all nodes that are used (transitively) by child/children.

        :return: list of nodes that are used
        :rtype: List[Node]
        """
        return self._child.nodes_used()

    def check(self):
        """Check predicate for consistency."""
        self._child.check()

    def requires_message(self):
        """Check whether any of child predicate(s) requires results of nodes (transitively).

        :return: True if the child requires results of parent task
        """
        return self._child.requires_message()


class AndPredicate(NaryPredicate):
    """And predicate representation."""

    def __str__(self):
        """Create a string representation of this predicate.

        :return: string representation (Python code)
        :rtype: str
        """
        return "(" + reduce(lambda x, y: str(x) + ' and ' + str(y), self._children) + ")"

    def ast(self):
        """Python AST of this predicate (construct transitively for all indirect children as well).

        :return: AST of describing all children predicates
        """
        return ast.BoolOp(ast.And(), [ast.Expr(value=x.ast()) for x in self._children])

    @staticmethod
    def create(tree, nodes_from, flow, can_inspect_results):
        """Create And predicate.

        :param tree: node from which should be predicate instantiated
        :type tree: List
        :param nodes_from: nodes that are used in described edge
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :param can_inspect_results: True if predicates in the condition can query task result
        :type can_inspect_results: bool
        :return: instance of cls
        """
        return NaryPredicate._create(tree, AndPredicate, nodes_from, flow, can_inspect_results)


class OrPredicate(NaryPredicate):
    """And predicate representation."""

    def __str__(self):
        """Create a string representation of this predicate.

        :return: string representation (Python code)
        :rtype: str
        """
        return "(" + reduce(lambda x, y: str(x) + ' or ' + str(y), self._children) + ")"

    def ast(self):
        """Python AST of this predicate (construct transitively for all indirect children as well).

        :return: AST of describing all children predicates
        """
        return ast.BoolOp(ast.Or(), [ast.Expr(value=x.ast()) for x in self._children])

    @staticmethod
    def create(tree, nodes_from, flow, can_inspect_results):
        """Create Or predicate.

        :param tree: node from which should be predicate instantiated
        :type tree: List
        :param nodes_from: nodes that are used in described edge
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :param can_inspect_results: True if predicates in the condition can query task result
        :type can_inspect_results: bool
        :return: instance of cls
        """
        return NaryPredicate._create(tree, OrPredicate, nodes_from, flow, can_inspect_results)


class NotPredicate(UnaryPredicate):
    """Unary or predicate representation."""

    def __str__(self):
        """Create a string representation of this predicate.

        :return: string representation (Python code)
        :rtype: str
        """
        return "(not %s)" % str(self._child)

    def ast(self):
        """Python AST of this predicate (construct transitively for all indirect children as well).

        :return: AST of describing all children predicates
        """
        return ast.UnaryOp(ast.Not(), ast.Expr(value=self._child.ast()))

    @staticmethod
    def create(tree, nodes_from, flow, can_inspect_results):
        """Create Or predicate.

        :param tree: node from which should be predicate instantiated
        :type tree: List
        :param nodes_from: nodes that are used in described edge
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :param can_inspect_results: True if predicates in the condition can query task result
        :type can_inspect_results: bool
        :return: instance of cls
        """
        return UnaryPredicate._create(tree, NotPredicate, nodes_from, flow, can_inspect_results)


class AlwaysTruePredicate(BuiltinPredicate):
    """Predicate used if condition in config file is omitted."""

    def __init__(self, flow):
        """Instantiate predicate that holds always True.

        :param flow: flow for which this predicate is used
        """
        super().__init__()
        self.flow = flow

    def __str__(self):
        """Create string representation of this predicate.

        :return: string representation (Python code)
        :rtype: str
        """
        return "True"

    def predicates_used(self):  # noqa
        return []

    def nodes_used(self):  # noqa
        return []

    def check(self):  # noqa
        """Check predicate for consistency."""

    def ast(self):
        """Python AST of this predicate (construct transitively for all indirect children as well).

        :return: AST of describing all children predicates
        """
        # We should return:
        #   return ast.NameConstant(value=True)
        # but it does not work with codegen
        return ast.Name(id='True', ctx=ast.Load())

    @staticmethod
    def create(tree, nodes_from, flow, can_inspect_results):  # noqa
        return AlwaysTruePredicate(flow)

    def requires_message(self):  # noqa
        return False
